#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import pathlib

import pytest

from PyInstaller.compat import is_darwin
from PyInstaller.utils.tests import importorskip, xfail

# Directory with testing modules used in some tests.
_MODULES_DIR = pathlib.Path(__file__).parent / 'modules'


# Test that PyiFrozenLoader.get_source() works as expected when source .py file is available.
def test_loader_get_source(pyi_builder):
    pyi_builder.test_source(
        """
        import pyi_dummy_module
        from pyimod02_importers import PyiFrozenLoader

        # Ensure the module is handled by PyiFrozenLoader.
        loader = pyi_dummy_module.__loader__
        assert isinstance(loader, PyiFrozenLoader)

        # Check that loader.get_source() returns source from the .py file.
        assert loader.get_source('pyi_dummy_module') is not None
        """,
        pyi_args=[
            '--add-data',
            f'{_MODULES_DIR / "pyi_dummy_module.py"}:.',
        ]
    )


def test_nameclash(pyi_builder):
    # test-case for issue #964: Nameclashes in module information gathering All pyinstaller specific module attributes
    # should be prefixed, to avoid nameclashes.
    pyi_builder.test_source("""
        import pyi_testmod_nameclash.nameclash
        """)


def test_relative_import(pyi_builder):
    pyi_builder.test_source(
        """
        import pyi_testmod_relimp.B.C
        from pyi_testmod_relimp.F import H
        import pyi_testmod_relimp.relimp1

        assert pyi_testmod_relimp.relimp1.name == 'pyi_testmod_relimp.relimp1'
        assert pyi_testmod_relimp.B.C.name == 'pyi_testmod_relimp.B.C'
        assert pyi_testmod_relimp.F.H.name == 'pyi_testmod_relimp.F.H'
        """
    )


def test_relative_import2(pyi_builder):
    pyi_builder.test_source(
        """
        import pyi_testmod_relimp2.bar
        import pyi_testmod_relimp2.bar.bar2

        pyi_testmod_relimp2.bar.say_hello_please()
        pyi_testmod_relimp2.bar.bar2.say_hello_please()
        """
    )


def test_relative_import3(pyi_builder):
    pyi_builder.test_source(
        """
        from pyi_testmod_relimp3a.aa import a1
        print(a1.getString())
        """
    )


@xfail(reason='modulegraph bug')
def test_import_missing_submodule(pyi_builder):
    # If a submodule is missing, the parent submodule must be imported.
    pyi_builder.test_source(
        """
        try:
            import pyi_testmod_missing_submod.aaa.bbb
        except ImportError as e:
            assert e.message.endswith(' bbb')
        else:
            raise RuntimeError('Buggy test-case: module pyi_testmod_missing_submod.aaa.bbb must not exist')
        # parent module exists and must be included
        __import__('pyi_testmod_missing_submod.aaa')
        """
    )


def test_import_submodule_global_shadowed(pyi_builder):
    """
    Functional test validating issue #1919.

    `ModuleGraph` previously ignored `from`-style imports of submodules from packages whose `__init__` submodules
    declared global variables of the same name as those submodules. This test exercises this sporadic edge case by
    unsuccessfully importing a submodule "shadowed" by a global variable of the same name defined by their package's
    `__init__` submodule.
    """

    pyi_builder.test_source(
        """
        # Assert that this submodule is shadowed by a string global variable.
        from pyi_testmod_submodule_global_shadowed import submodule
        assert type(submodule) == str

        # Assert that this submodule is still frozen into this test application.
        # To do so:
        #
        # 1. Delete this global variable from its parent package.
        # 2. Assert that this submodule is unshadowed by this global variable.
        import pyi_testmod_submodule_global_shadowed, sys
        del  pyi_testmod_submodule_global_shadowed.submodule
        from pyi_testmod_submodule_global_shadowed import submodule
        assert type(submodule) == type(sys)
        """
    )


def test_import_submodule_global_unshadowed(pyi_builder):
    """
    Functional test validating issue #1919.

    `ModuleGraph` previously ignored `from`-style imports of submodules from packages whose `__init__` submodules
    declared global variables of the same name as those submodules. This test exercises this sporadic edge case by
    successfully importing a submodule:

    * Initially "shadowed" by a global variable of the same name defined by their package's `__init__` submodule.
    * Subsequently "unshadowed" when this global variable is then undefined by their package's `__init__` submodule.
    """

    pyi_builder.test_source(
        """
        # Assert that this submodule is unshadowed by this global variable.
        import sys
        from pyi_testmod_submodule_global_unshadowed import submodule
        assert type(submodule) == type(sys)
        """
    )


def test_import_submodule_from_aliased_pkg(pyi_builder, script_dir):
    pyi_builder.test_source(
        """
        import sys
        import pyi_testmod_submodule_from_aliased_pkg

        sys.modules['alias_name'] = pyi_testmod_submodule_from_aliased_pkg

        from alias_name import submodule
        """, ['--additional-hooks-dir', f"{script_dir / 'pyi_hooks'}"]
    )


def test_module_with_coding_utf8(pyi_builder):
    # Module ``utf8_encoded_module`` simply has an ``coding`` header and uses same German umlauts.
    pyi_builder.test_source("import module_with_coding_utf8")


# Test that our PyiFrozenLoader's get_source() method can load source files with utf-8 emoji characters.
# See issue #6143.
def test_source_utf8_emoji(pyi_builder):
    # Collect the module's source as data file
    add_data_arg = f"{_MODULES_DIR / 'module_with_utf8_emoji.py'}:."
    pyi_builder.test_source(
        """
        import inspect

        import module_with_utf8_emoji

        # Retrieve source code
        source = inspect.getsource(module_with_utf8_emoji)
        """, ['--add-data', add_data_arg]
    )


def test_hiddenimport(pyi_builder):
    # The script simply does nothing, not even print out a line. The check is done by comparing with
    # logs/test_hiddenimport.toc
    pyi_builder.test_source('pass', ['--hidden-import=a_hidden_import'])


def test_error_during_import(pyi_builder):
    # See ticket #27: historically, PyInstaller was catching all errors during imports...
    pyi_builder.test_source(
        """
        try:
            import error_during_import2
        except KeyError:
            print("OK")
        else:
            raise RuntimeError("failure!")
        """
    )


def test_import_non_existing_raises_import_error(pyi_builder):
    pyi_builder.test_source(
        """
        try:
            import zzzzzz.zzzzzzzz.zzzzzzz.non.existing.module.error_during_import2
        except ImportError:
            print("OK")
        else:
            raise RuntimeError("ImportError not raised")
        """
    )


# Verify that __path__ is respected for imports from the filesystem:
#
# * pyi_testmod_path/
#
#   * __init__.py -- inserts a/ into __path__, then imports b, which now refers to a/b.py, not ./b.py.
#   * b.py - raises an exception. **Should not be imported.**
#   * a/ -- contains no __init__.py.
#
#     * b.py - Empty. Should be imported.
@xfail(reason='__path__ not respected for filesystem modules.')
def test_import_respects_path(pyi_builder, script_dir):
    pyi_builder.test_source(
        'import pyi_testmod_path',
        pyi_args=['--additional-hooks-dir', str(script_dir / 'pyi_hooks')],
    )


# Verify correct handling of sys.meta_path redirects like pkg_resources 28.6.1 does: '_vendor.xxx' gets imported as
# 'extern.xxx' and using '__import__()'. Note: This also requires a hook, since 'pyi_testmod_metapath1._vendor' is not
# imported directly and won't be found by modulegraph.
def test_import_metapath1(pyi_builder, script_dir):
    pyi_builder.test_source(
        'import pyi_testmod_metapath1',
        pyi_args=['--additional-hooks-dir', str(script_dir / 'pyi_hooks')],
    )


@importorskip('PyQt5')
def test_import_pyqt5_uic_port(script_dir, pyi_builder):
    extra_path = _MODULES_DIR / 'pyi_import_pyqt_uic_port'
    pyi_builder.test_script(
        'pyi_import_pyqt5_uic_port.py',
        # Add the path to a fake PyQt5 package, used for this test.
        pyi_args=['--path', str(extra_path)]
    )


# Check whether modulegraph finds pyi_splash
def test_import_pyi_splash(pyi_builder):
    pyi_builder.test_source(
        """
        import pyi_splash
        assert hasattr(pyi_splash, "_initialized")
        """
    )


#--- unzipped .egg support ----


def test_egg_unzipped(pyi_builder):
    pathex = _MODULES_DIR / 'pyi_test_egg' / 'pyi_egg_unzipped.egg'
    hooks_dir = _MODULES_DIR / 'pyi_test_egg' / 'hooks'
    pyi_builder.test_source(
        """
        # This code is part of the package for testing eggs in `PyInstaller`.
        import os
        import pkg_resources

        # Test ability to load resource.
        expected_data = 'This is data file for `unzipped`.'.encode('ascii')
        t = pkg_resources.resource_string('unzipped_egg', 'data/datafile.txt')
        print('Resource: %s' % t)
        t_filename = pkg_resources.resource_filename('unzipped_egg', 'data/datafile.txt')
        print('Resource filename: %s' % t_filename)
        assert t.rstrip() == expected_data

        # Test ability that module from .egg is able to load resource.
        import unzipped_egg
        assert unzipped_egg.data == expected_data

        print('Okay.')
        """,
        pyi_args=['--paths', str(pathex), '--additional-hooks-dir', str(hooks_dir)],
    )  # yapf: disable


def test_egg_unzipped_metadata_pkg_resources(pyi_builder):
    pathex = _MODULES_DIR / 'pyi_test_egg' / 'pyi_egg_unzipped.egg'
    hooks_dir = _MODULES_DIR / 'pyi_test_egg' / 'hooks'
    pyi_builder.test_source(
        """
        import pkg_resources

        # Metadata should be automatically collected due to pkg_resources.get_distribution() call with literal argument
        # (which is picked up by PyInstaller's bytecode analysis).
        dist = pkg_resources.get_distribution('pyi_egg_unzipped')
        print(f"dist: {dist!r}")

        # Version is taken from metadata
        assert dist.version == '0.1', f"Unexpected version {dist.version!r}"
        # Project name is taken from egg name
        assert dist.project_name == 'pyi-egg-unzipped', f"Unexpected project name {dist.project_name!r}"
        """,
        pyi_args=['--paths', str(pathex), '--additional-hooks-dir', str(hooks_dir)],
    )  # yapf: disable


def test_egg_unzipped_metadata_importlib_metadata(pyi_builder):
    pathex = _MODULES_DIR / 'pyi_test_egg' / 'pyi_egg_unzipped.egg'
    hooks_dir = _MODULES_DIR / 'pyi_test_egg' / 'hooks'
    pyi_builder.test_source(
        """
        try:
            import importlib_metadata
        except ModuleNotFoundError:
            import importlib.metadata as importlib_metadata

        # Metadata should be automatically collected due to importlib_metadata.version() call with literal argument
        # (which is picked up by PyInstaller's bytecode analysis).
        version = importlib_metadata.version('pyi_egg_unzipped')
        print(f"version: {version!r}")
        assert version == '0.1', f"Unexpected version {version!r}"

        # NOTE: in contrast to pkg_resources, importlib_metadata seems to read the name from metadata instead of
        # deriving it from egg directory name.
        metadata = importlib_metadata.metadata('pyi_egg_unzipped')
        print(f"metadata: {metadata!r}")
        assert metadata['Name'] == 'unzipped-egg', f"Unexpected Name {metadata['Name']!r}"
        assert metadata['Version'] == '0.1', f"Unexpected Version {metadata['Version']!r}"

        dist = importlib_metadata.distribution('pyi_egg_unzipped')
        print(f"dist: {dist!r}")
        assert dist.name == 'unzipped-egg', f"Unexpected name {dist.name!r}"
        assert dist.version == '0.1', f"Unexpected version {dist.version!r}"
        """,
        pyi_args=['--paths', str(pathex), '--additional-hooks-dir', str(hooks_dir)],
    )  # yapf: disable


#--- namespaces ---


def test_nspkg1(pyi_builder):
    # Test inclusion of namespace packages implemented using pkg_resources.declare_namespace
    extra_paths = (_MODULES_DIR / 'nspkg1-pkg').glob('*.egg')
    paths_arg = os.pathsep.join([str(path) for path in extra_paths])
    pyi_builder.test_source(
        """
        import nspkg1.aaa
        import nspkg1.bbb.zzz
        import nspkg1.ccc
        """,
        pyi_args=['--paths', paths_arg],
    )


def test_nspkg1_empty(pyi_builder):
    # Test inclusion of a namespace-only packages in an zipped egg. This package only defines the namespace, nothing is
    # contained there.
    extra_paths = (_MODULES_DIR / 'nspkg1-pkg').glob('*.egg')
    paths_arg = os.pathsep.join([str(path) for path in extra_paths])
    pyi_builder.test_source(
        """
        import nspkg1
        print (nspkg1)
        """,
        pyi_args=['--paths', paths_arg],
    )


def test_nspkg1_bbb_zzz(pyi_builder):
    # Test inclusion of a namespace packages in an zipped egg
    extra_paths = (_MODULES_DIR / 'nspkg1-pkg').glob('*.egg')
    paths_arg = os.pathsep.join([str(path) for path in extra_paths])
    pyi_builder.test_source(
        """
        import nspkg1.bbb.zzz
        """,
        pyi_args=['--paths', paths_arg],
    )


def test_nspkg2(pyi_builder):
    # Test inclusion of namespace packages implemented as nspkg.pth-files
    extra_paths = [_MODULES_DIR / 'nspkg2-pkg']
    paths_arg = os.pathsep.join([str(path) for path in extra_paths])
    pyi_builder.test_source(
        """
        import nspkg2.aaa
        import nspkg2.bbb.zzz
        import nspkg2.ccc
        """,
        pyi_args=['--paths', paths_arg],
    )


@xfail(reason="modulegraph implements `pkgutil.extend_path` wrong")
def test_nspkg3(pyi_builder):
    extra_paths = (_MODULES_DIR / 'nspkg3-pkg').glob('*.egg')
    paths_arg = os.pathsep.join([str(path) for path in extra_paths])
    pyi_builder.test_source(
        """
        import nspkg3.aaa
        try:
            # pkgutil ignores items of sys.path that are not strings referring to existing directories. So this zipped
            # egg *must* be ignored.
            import nspkg3.bbb.zzz
        except ImportError:
            pass
        else:
            raise SystemExit('nspkg3.bbb.zzz found but should not')
        try:
            import nspkg3.a
        except ImportError:
            pass
        else:
            raise SystemExit('nspkg3.a found but should not')
        import nspkg3.ccc
        """,
        pyi_args=['--paths', paths_arg],
    )


def test_nspkg3_empty(pyi_builder):
    # Test inclusion of a namespace-only package in a zipped egg using pkgutil.extend_path. This package only defines
    # namespace, nothing is contained there.
    extra_paths = (_MODULES_DIR / 'nspkg3-pkg').glob('*_empty.egg')
    paths_arg = os.pathsep.join([str(path) for path in extra_paths])
    pyi_builder.test_source(
        """
        import nspkg3
        print (nspkg3)
        """,
        pyi_args=['--paths', paths_arg],
    )


def test_nspkg3_aaa(pyi_builder):
    # Test inclusion of a namespace package in an directory using pkgutil.extend_path
    extra_paths = (_MODULES_DIR / 'nspkg3-pkg').glob('*.egg')
    paths_arg = os.pathsep.join([str(path) for path in extra_paths])
    pyi_builder.test_source(
        """
        import nspkg3.aaa
        """,
        pyi_args=['--paths', paths_arg],
    )


def test_nspkg3_bbb_zzz(pyi_builder):
    # Test inclusion of a namespace package in an zipped egg using pkgutil.extend_path
    extra_paths = (_MODULES_DIR / 'nspkg3-pkg').glob('*.egg')
    paths_arg = os.pathsep.join([str(path) for path in extra_paths])
    pyi_builder.test_source(
        """
        import nspkg3.bbb.zzz
        """,
        pyi_args=['--paths', paths_arg],
    )


def test_nspkg_pep420(pyi_builder):
    # Test inclusion of PEP 420 namespace packages.
    extra_paths = (_MODULES_DIR / 'nspkg-pep420').glob('path*')
    paths_arg = os.pathsep.join([str(path) for path in extra_paths])
    pyi_builder.test_source(
        """
        import package.sub1
        import package.sub2
        import package.subpackage.sub
        import package.nspkg.mod
        """,
        pyi_args=['--paths', paths_arg],
    )


def test_nspkg_attributes(pyi_builder):
    # Test that non-PEP-420 namespace packages (e.g., the ones using
    # pkg_resources.declare_namespace) have proper attributes:
    #  * __path__ attribute should contain at least one path
    #  * __file__ attribute should point to an __init__ file within __path__
    extra_paths = (_MODULES_DIR / 'nspkg1-pkg').glob('*.egg')
    paths_arg = os.pathsep.join([str(path) for path in extra_paths])
    pyi_builder.test_source(
        """
        import os
        import nspkg1

        def validate_nspkg(pkg):
            from sys import version_info
            # Validate __path__
            path = getattr(pkg, '__path__', None)
            assert path is not None and len(path) >= 1, "invalid __path__"
            # Validate __file__
            file = pkg.__file__
            assert os.path.dirname(file) in path, \
                "dirname(__file__) does not point to __path__"
            assert os.path.basename(file).startswith('__init__.'), \
                "basename(__file__) does not start with __init__.!"

        validate_nspkg(nspkg1)
        """,
        pyi_args=['--paths', paths_arg],
    )


def test_nspkg_attributes_pep420(pyi_builder):
    # Test that PEP-420 namespace packages have proper attributes:
    #  * __path__ should contain at least one path
    #  * __file__ should be None
    extra_paths = (_MODULES_DIR / 'nspkg-pep420').glob('path*')
    paths_arg = os.pathsep.join([str(path) for path in extra_paths])
    pyi_builder.test_source(
        """
        import package
        import package.nspkg

        def validate_nspkg_pep420(pkg):
            from sys import version_info
            # Validate __path__
            path = getattr(pkg, '__path__', None)
            assert path is not None and len(path) >= 1, "invalid __path__"
            # Validate __file__
            assert getattr(pkg, '__file__') is None, "invalid __file__"

        validate_nspkg_pep420(package)
        validate_nspkg_pep420(package.nspkg)
        """,
        pyi_args=['--paths', paths_arg],
    )


#--- hooks related stuff ---


# imp leaks file handles.
@pytest.mark.filterwarnings("ignore", category=ResourceWarning)
def test_pkg_without_hook_for_pkg(pyi_builder, script_dir):
    # The package `pkg_without_hook_for_pkg` does not have a hook, but `pkg_without_hook_for_pkg.sub1` has one. And this
    # hook includes the "hidden" import `pkg_without_hook_for_pkg.sub1.sub11`
    pyi_builder.test_source(
        'import pkg_without_hook_for_pkg.sub1',
        pyi_args=['--additional-hooks-dir', str(script_dir / 'pyi_hooks')],
    )


def test_app_with_plugin(pyi_builder, data_dir):
    add_data_arg = f"{data_dir / 'static_plugin.py'}:."
    pyi_builder.test_script(
        'pyi_app_with_plugin.py',
        pyi_args=['--add-data', add_data_arg],
    )


def test_app_has_moved_error(pyi_builder, tmp_path):
    """
    Test graceful exit from the user moving/deleting the application whilst it's still running.
    """
    pyi_builder.test_source(
        f"""
        import os
        import sys
        os.rename(sys.executable, {repr(str(tmp_path / "something-else"))})
        try:
            # Import some non-builtin module which hasn't already been loaded.
            import csv
        except SystemExit:
            pass
        else:
            assert 0, "A system exit should have been raised."
        """
    )


def test_package_with_mixed_collection_mode(pyi_builder):
    # Test that PyInstaller's frozen importer and python's own `_frozen_importlib_external.PathFinder` complement, and
    # not exclude, each other. This is pre-requisite for having pure-python modules collected in PYZ archive, while
    # binary extensions (and some pure-python modules as well, if necessary) are collected as separate.
    pathex = _MODULES_DIR / 'pyi_mixed_collection_mode' / 'modules'
    hooks_dir = _MODULES_DIR / 'pyi_mixed_collection_mode' / 'hooks'
    pyi_builder.test_source(
        """
        import mypackage
        print(mypackage.a)
        print(mypackage.b)
        """,
        pyi_args=['--paths', str(pathex), '--additional-hooks-dir', str(hooks_dir)],
    )  # yapf: disable


# Tests for run-time sys.path modifications, typically with aim of exposing some part of a package to the outside world
# as a top-level package; for example, to expose a vendored package to the outside world. These tests aim to verify that
# we can handle dynamic sys.path modifications within PYZ-collected packages and honor the order of entries in sys.path.


# Shared implementation
def _test_sys_path_with_vendored_package(pyi_builder, modification_type, expected_string, extra_pyi_args=None):
    pathex = _MODULES_DIR / 'pyi_sys_path_with_vendored_package'

    pyi_args = [
        '--paths', str(pathex),
        '--hiddenimport', 'myotherpackage._vendored.mypackage.mod',
    ]  # yapf: disable

    if extra_pyi_args:
        pyi_args += extra_pyi_args

    # On macOS, also build and test .app bundle executable
    if is_darwin:
        pyi_args += ['--windowed']

    pyi_builder.test_source(
        f"""
        import myotherpackage
        myotherpackage.setup_vendored_packages('{modification_type}')

        import mypackage
        secret = mypackage.get_secret_string()
        assert secret == '{expected_string}', f"Unexpected secret string: {{secret!r}}"
        """,
        pyi_args=pyi_args,
    )


# In this scenario, we have only vendored package available - we intentionally suppress collection of stand-alone
# package during the build.
def test_sys_path_with_vendored_package_no_standalone(pyi_builder):
    _test_sys_path_with_vendored_package(pyi_builder, "append", "vendored", ["--exclude", "mypackage"])


# In this scenario, we have both stand-alone and vendored package available in sys.path; vendored package directory is
# appended to sys.path, so we expect to import the stand-alone version.
def test_sys_path_with_vendored_package_append(pyi_builder):
    _test_sys_path_with_vendored_package(pyi_builder, "append", "standalone")


# In this scenario, we have both stand-alone and vendored package available in sys.path; vendored package directory is
# prepended to sys.path, so we expect to import the vendored version.
def test_sys_path_with_vendored_package_prepend(pyi_builder):
    _test_sys_path_with_vendored_package(pyi_builder, "prepend", "vendored")


# Tests for run-time sys.path modifications that result in a PEP420 namespace package being split across different
# locations, both within the PYZ archive and in location external to frozen application.


# The test supports two import orders: standalone part, vendored part, external part; and reverse. Both orders are
# tested in case it matters whether the namespace package is first discovered by PyInstaller's frozen importer or by
# python's `_frozen_importlib_external.PathFinder`.
# Additionally, two sys.path modification strategies are tested: adding each new entry just before we import the
# corresponding module, or adding all entries in advance. Adding entries one by one should trigger recomputation
# of the `_NamespacePath` object on each subsequent import.
@pytest.mark.parametrize('import_order', ['forward', 'reverse'])
@pytest.mark.parametrize('path_modification', ['one_by_one', 'all_in_advance'])
def test_split_location_pep420_namespace_package(pyi_builder, import_order, path_modification):
    modules_root = _MODULES_DIR / 'pyi_split_location_pep420_namespace_package'

    pyi_args = [
        '--paths', str(modules_root / 'modules'),
        '--additional-hooks-dir', str(modules_root / 'hooks'),
        '--hiddenimport', 'myotherpackage._vendored.mynamespacepackage.vendored_pyz',
        '--hiddenimport', 'myotherpackage._vendored.mynamespacepackage.vendored_py',
    ]  # yapf: disable

    # On macOS, also build and test .app bundle executable
    if is_darwin:
        pyi_args += ['--windowed']

    # Path to external part needs to be passed to program via command-line arguments
    app_args = [os.path.join(modules_root, 'external-location')]

    # Test programs for all four combinations
    _test_programs = {}
    _test_programs[('forward', 'one_by_one')] = \
        """
        import sys

        external_path = sys.argv[1]
        from myotherpackage import vendored_path

        import mynamespacepackage.standalone_pyz
        print(mynamespacepackage.standalone_pyz)

        import mynamespacepackage.standalone_py
        print(mynamespacepackage.standalone_py)

        sys.path.append(vendored_path)
        import mynamespacepackage.vendored_pyz
        print(mynamespacepackage.vendored_pyz)

        import mynamespacepackage.vendored_py
        print(mynamespacepackage.vendored_py)

        sys.path.append(external_path)
        import mynamespacepackage.external_py
        print(mynamespacepackage.external_py)
        """
    _test_programs[('forward', 'all_in_advance')] = \
        """
        import sys

        external_path = sys.argv[1]
        from myotherpackage import vendored_path

        sys.path.append(vendored_path)
        sys.path.append(external_path)

        import mynamespacepackage.standalone_pyz
        print(mynamespacepackage.standalone_pyz)

        import mynamespacepackage.standalone_py
        print(mynamespacepackage.standalone_py)

        import mynamespacepackage.vendored_pyz
        print(mynamespacepackage.vendored_pyz)

        import mynamespacepackage.vendored_py
        print(mynamespacepackage.vendored_py)

        import mynamespacepackage.external_py
        print(mynamespacepackage.external_py)
        """
    _test_programs[('reverse', 'one_by_one')] = \
        """
        import sys

        external_path = sys.argv[1]
        from myotherpackage import vendored_path

        sys.path.insert(0, external_path)
        import mynamespacepackage.external_py
        print(mynamespacepackage.external_py)

        sys.path.insert(1, vendored_path)
        import mynamespacepackage.vendored_py
        print(mynamespacepackage.vendored_py)

        import mynamespacepackage.vendored_pyz
        print(mynamespacepackage.vendored_pyz)

        import mynamespacepackage.standalone_py
        print(mynamespacepackage.standalone_py)

        import mynamespacepackage.standalone_pyz
        print(mynamespacepackage.standalone_pyz)
        """
    _test_programs[('reverse', 'all_in_advance')] = \
        """
        import sys

        external_path = sys.argv[1]
        from myotherpackage import vendored_path

        sys.path.insert(0, external_path)
        sys.path.insert(1, vendored_path)

        import mynamespacepackage.external_py
        print(mynamespacepackage.external_py)

        import mynamespacepackage.vendored_py
        print(mynamespacepackage.vendored_py)

        import mynamespacepackage.vendored_pyz
        print(mynamespacepackage.vendored_pyz)

        import mynamespacepackage.standalone_py
        print(mynamespacepackage.standalone_py)

        import mynamespacepackage.standalone_pyz
        print(mynamespacepackage.standalone_pyz)
        """

    # Build and run the appropriate test program
    pyi_builder.test_source(
        _test_programs[(import_order, path_modification)],
        pyi_args=pyi_args,
        app_args=app_args,
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
PEP-302 and PEP-451 importers for frozen applications.
"""

# **NOTE** This module is used during bootstrap.
# Import *ONLY* builtin modules or modules that are collected into the base_library.zip archive.
# List of built-in modules: sys.builtin_module_names
# List of modules collected into base_library.zip: PyInstaller.compat.PY3_BASE_MODULES

import sys
import os
import io

import _frozen_importlib
import _thread

import pyimod01_archive

if sys.flags.verbose and sys.stderr:

    def trace(msg, *a):
        sys.stderr.write(msg % a)
        sys.stderr.write("\n")
else:

    def trace(msg, *a):
        pass


def _decode_source(source_bytes):
    """
    Decode bytes representing source code and return the string. Universal newline support is used in the decoding.
    Based on CPython's implementation of the same functionality:
    https://github.com/python/cpython/blob/3.9/Lib/importlib/_bootstrap_external.py#L679-L688
    """
    # Local import to avoid including `tokenize` and its dependencies in `base_library.zip`
    from tokenize import detect_encoding
    source_bytes_readline = io.BytesIO(source_bytes).readline
    encoding = detect_encoding(source_bytes_readline)
    newline_decoder = io.IncrementalNewlineDecoder(decoder=None, translate=True)
    return newline_decoder.decode(source_bytes.decode(encoding[0]))


# Global instance of PYZ archive reader. Initialized by install().
pyz_archive = None

# Some runtime hooks might need to traverse available frozen package/module hierarchy to simulate filesystem.
# Such traversals can be efficiently implemented using a prefix tree (trie), whose computation we defer until first
# access.
_pyz_tree_lock = _thread.RLock()
_pyz_tree = None


def get_pyz_toc_tree():
    global _pyz_tree

    with _pyz_tree_lock:
        if _pyz_tree is None:
            _pyz_tree = _build_pyz_prefix_tree(pyz_archive)
        return _pyz_tree


# Populate list of unresolved (original) and resolved paths to top-level directory, used when trying to determine
# relative path.
_TOP_LEVEL_DIRECTORY_PATHS = []

# Original sys._MEIPASS value; ensure separators are normalized (e.g., when using msys2 python).
_TOP_LEVEL_DIRECTORY = os.path.normpath(sys._MEIPASS)
_TOP_LEVEL_DIRECTORY_PATHS.append(_TOP_LEVEL_DIRECTORY)

# Fully resolve sys._MEIPASS in case its location is symlinked at some level; for example, system temporary directory
# (used by onefile builds) is usually a symbolic link under macOS.
_RESOLVED_TOP_LEVEL_DIRECTORY = os.path.realpath(_TOP_LEVEL_DIRECTORY)
if os.path.normcase(_RESOLVED_TOP_LEVEL_DIRECTORY) != os.path.normcase(_TOP_LEVEL_DIRECTORY):
    _TOP_LEVEL_DIRECTORY_PATHS.append(_RESOLVED_TOP_LEVEL_DIRECTORY)

# If we are running as macOS .app bundle, compute the alternative top-level directory path as well.
_is_macos_app_bundle = False
if sys.platform == 'darwin' and _TOP_LEVEL_DIRECTORY.endswith("Contents/Frameworks"):
    _is_macos_app_bundle = True

    _ALTERNATIVE_TOP_LEVEL_DIRECTORY = os.path.join(
        os.path.dirname(_TOP_LEVEL_DIRECTORY),
        'Resources',
    )
    _TOP_LEVEL_DIRECTORY_PATHS.append(_ALTERNATIVE_TOP_LEVEL_DIRECTORY)

    _RESOLVED_ALTERNATIVE_TOP_LEVEL_DIRECTORY = os.path.join(
        os.path.dirname(_RESOLVED_TOP_LEVEL_DIRECTORY),
        'Resources',
    )
    if _RESOLVED_ALTERNATIVE_TOP_LEVEL_DIRECTORY != _ALTERNATIVE_TOP_LEVEL_DIRECTORY:
        _TOP_LEVEL_DIRECTORY_PATHS.append(_RESOLVED_ALTERNATIVE_TOP_LEVEL_DIRECTORY)


# Helper for computing PYZ prefix tree
def _build_pyz_prefix_tree(pyz_archive):
    tree = dict()
    for entry_name, entry_data in pyz_archive.toc.items():
        name_components = entry_name.split('.')
        typecode = entry_data[0]
        current = tree
        if typecode in {pyimod01_archive.PYZ_ITEM_PKG, pyimod01_archive.PYZ_ITEM_NSPKG}:
            # Package; create new dictionary node for its modules
            for name_component in name_components:
                current = current.setdefault(name_component, {})
        else:
            # Module; create the leaf node (empty string)
            for name_component in name_components[:-1]:
                current = current.setdefault(name_component, {})
            current[name_components[-1]] = ''
    return tree


class PyiFrozenFinder:
    """
    PyInstaller's frozen path entry finder for specific search path.

    Per-path instances allow us to properly translate the given module name ("fullname") into full PYZ entry name.
    For example, with search path being `sys._MEIPASS`, the module "mypackage.mod" would translate to "mypackage.mod"
    in the PYZ archive. However, if search path was `sys._MEIPASS/myotherpackage/_vendored` (for example, if
    `myotherpacakge` added this path to `sys.path`), then "mypackage.mod" would need to translate to
    "myotherpackage._vendored.mypackage.mod" in the PYZ archive.
    """
    def __repr__(self):
        return f"{self.__class__.__name__}({self._path})"

    @classmethod
    def path_hook(cls, path):
        trace(f"PyInstaller: running path finder hook for path: {path!r}")
        try:
            finder = cls(path)
            trace("PyInstaller: hook succeeded")
            return finder
        except Exception as e:
            trace(f"PyInstaller: hook failed: {e}")
            raise

    def __init__(self, path):
        self._path = path  # Store original path, as given.
        self._pyz_archive = pyz_archive

        # Compute relative path to the top-level application directory. Do not try to resolve the path itself, because
        # it might contain symbolic links in parts other than the prefix that corresponds to the top-level application
        # directory. See #8994 for an example (files symlinked from a common directory outside of the top-level
        # application directory). Instead, try to compute relative path w.r.t. the original and the resolved top-level
        # application directory.
        for top_level_path in _TOP_LEVEL_DIRECTORY_PATHS:
            try:
                relative_path = os.path.relpath(path, top_level_path)
            except ValueError:
                continue  # Failed to compute relative path w.r.t. the given top-level directory path.

            if relative_path.startswith('..'):
                continue  # Relative path points outside of the given top-level directory.

            break  # Successful match; stop here.
        else:
            raise ImportError("Failed to determine relative path w.r.t. top-level application directory.")

        # Ensure that path does not point to a file on filesystem. Strictly speaking, we should be checking that the
        # given path is a valid directory, but that would need to check both PYZ and filesystem. So for now, limit the
        # check to catch paths pointing to file, because that breaks `runpy.run_path()`, as per #8767.
        if os.path.isfile(path):
            raise ImportError("only directories are supported")

        if relative_path == '.':
            self._pyz_entry_prefix = ''
        else:
            self._pyz_entry_prefix = '.'.join(relative_path.split(os.path.sep))

    def _compute_pyz_entry_name(self, fullname):
        """
        Convert module fullname into PYZ entry name, subject to the prefix implied by this finder's search path.
        """
        tail_module = fullname.rpartition('.')[2]

        if self._pyz_entry_prefix:
            return self._pyz_entry_prefix + "." + tail_module
        else:
            return tail_module

    @property
    def fallback_finder(self):
        """
        Opportunistically create a *fallback finder* using `sys.path_hooks` entries that are located *after* our hook.
        The main goal of this exercise is to obtain an instance of python's FileFinder, but in theory any other hook
        that comes after ours is eligible to be a fallback.

        Having this fallback allows our finder to "cooperate" with python's FileFinder, as if the two were a single
        finder, which allows us to work around the python's PathFinder permitting only one finder instance per path
        without subclassing FileFinder.
        """
        if hasattr(self, '_fallback_finder'):
            return self._fallback_finder

        # Try to instantiate fallback finder
        our_hook_found = False

        self._fallback_finder = None
        for idx, hook in enumerate(sys.path_hooks):
            if hook == self.path_hook:
                our_hook_found = True
                continue  # Our hook

            if not our_hook_found:
                continue  # Skip hooks before our hook

            try:
                self._fallback_finder = hook(self._path)
                break
            except ImportError:
                pass

        return self._fallback_finder

    def _find_fallback_spec(self, fullname, target):
        """
        Attempt to find the spec using fallback finder, which is opportunistically created here. Typically, this would
        be python's FileFinder, which can discover specs for on-filesystem modules, such as extension modules and
        modules that are collected only as source .py files.

        Having this fallback allows our finder to "cooperate" with python's FileFinder, as if the two were a single
        finder, which allows us to work around the python's PathFinder permitting only one finder instance per path
        without subclassing FileFinder.
        """
        if not hasattr(self, '_fallback_finder'):
            self._fallback_finder = self._get_fallback_finder()

        if self._fallback_finder is None:
            return None

        return self._fallback_finder.find_spec(fullname, target)

    #-- Core PEP451 finder functionality, modeled after importlib.abc.PathEntryFinder
    # https://docs.python.org/3/library/importlib.html#importlib.abc.PathEntryFinder
    def invalidate_caches(self):
        """
        A method which, when called, should invalidate any internal cache used by the finder. Used by
        importlib.invalidate_caches() when invalidating the caches of all finders on sys.meta_path.

        https://docs.python.org/3/library/importlib.html#importlib.abc.MetaPathFinder.invalidate_caches
        """
        # We do not use any caches, but if we have created a fallback finder, propagate the function call.
        # NOTE: use getattr() with _fallback_finder attribute, in order to avoid unnecessary creation of the
        # fallback finder in case when it does not exist yet.
        fallback_finder = getattr(self, '_fallback_finder', None)
        if fallback_finder is not None:
            if hasattr(fallback_finder, 'invalidate_caches'):
                fallback_finder.invalidate_caches()

    def find_spec(self, fullname, target=None):
        """
        A method for finding a spec for the specified module. The finder will search for the module only within the
        path entry to which it is assigned. If a spec cannot be found, None is returned. When passed in, target is a
        module object that the finder may use to make a more educated guess about what spec to return.

        https://docs.python.org/3/library/importlib.html#importlib.abc.PathEntryFinder.find_spec
        """
        trace(f"{self}: find_spec: called with fullname={fullname!r}, target={fullname!r}")

        # Convert fullname to PYZ entry name.
        pyz_entry_name = self._compute_pyz_entry_name(fullname)

        # Try looking up the entry in the PYZ archive
        entry_data = self._pyz_archive.toc.get(pyz_entry_name)
        if entry_data is None:
            # Entry not found - try using fallback finder (for example, python's own FileFinder) to resolve on-disk
            # resources, such as extension modules and modules that are collected only as source .py files.
            trace(f"{self}: find_spec: {fullname!r} not found in PYZ...")

            if self.fallback_finder is not None:
                trace(f"{self}: find_spec: attempting resolve using fallback finder {self.fallback_finder!r}.")
                fallback_spec = self.fallback_finder.find_spec(fullname, target)
                trace(f"{self}: find_spec: fallback finder returned spec: {fallback_spec!r}.")
                return fallback_spec
            else:
                trace(f"{self}: find_spec: fallback finder is not available.")

            return None

        # Entry found
        typecode = entry_data[0]
        trace(f"{self}: find_spec: found {fullname!r} in PYZ as {pyz_entry_name!r}, typecode={typecode}")

        if typecode == pyimod01_archive.PYZ_ITEM_NSPKG:
            # PEP420 namespace package
            # We can use regular list for submodule_search_locations; the caller (i.e., python's PathFinder) takes care
            # of constructing _NamespacePath from it.
            spec = _frozen_importlib.ModuleSpec(fullname, None)
            spec.submodule_search_locations = [
                # NOTE: since we are using sys._MEIPASS as prefix, we need to construct path from resolved PYZ entry
                # name (equivalently, we could combine `self._path` and last part of `fullname`).
                os.path.join(sys._MEIPASS, pyz_entry_name.replace('.', os.path.sep)),
            ]
            return spec

        is_package = typecode == pyimod01_archive.PYZ_ITEM_PKG

        # Instantiate frozen loader for the module
        loader = PyiFrozenLoader(
            name=fullname,
            pyz_archive=self._pyz_archive,
            pyz_entry_name=pyz_entry_name,
            is_package=is_package,
        )

        # Resolve full filename, as if the module/package was located on filesystem. This is done by the loader.
        origin = loader.path

        # Construct spec for module, using all collected information.
        spec = _frozen_importlib.ModuleSpec(
            fullname,
            loader,
            is_package=is_package,
            origin=origin,
        )

        # Make the import machinery set __file__.
        # PEP 451 says: "has_location" is true if the module is locatable. In that case the spec's origin is used
        # as the location and __file__ is set to spec.origin. If additional location information is required
        # (e.g., zipimport), that information may be stored in spec.loader_state.
        spec.has_location = True

        # Set submodule_search_locations for packages. Seems to be required for importlib_resources from 3.2.0;
        # see issue #5395.
        if is_package:
            spec.submodule_search_locations = [os.path.dirname(origin)]

        return spec

    # The following methods are part of legacy PEP302 finder interface. They have been deprecated since python 3.4,
    # and removed in python 3.12. Provide compatibility shims to accommodate code that might still be using them.
    if sys.version_info[:2] < (3, 12):

        def find_loader(self, fullname):
            """
            A legacy method for finding a loader for the specified module. Returns a 2-tuple of (loader, portion) where
            portion is a sequence of file system locations contributing to part of a namespace package. The loader may
            be None while specifying portion to signify the contribution of the file system locations to a namespace
            package. An empty list can be used for portion to signify the loader is not part of a namespace package. If
            loader is None and portion is the empty list then no loader or location for a namespace package were found
            (i.e. failure to find anything for the module).

            Deprecated since python 3.4, removed in 3.12.
            """
            # Based on:
            # https://github.com/python/cpython/blob/v3.11.9/Lib/importlib/_bootstrap_external.py#L1587-L1600
            spec = self.find_spec(fullname)
            if spec is None:
                return None, []
            return spec.loader, spec.submodule_search_locations or []

        def find_module(self, fullname):
            """
            A concrete implementation of Finder.find_module() which is equivalent to self.find_loader(fullname)[0].

            Deprecated since python 3.4, removed in 3.12.
            """
            # Based on:
            # https://github.com/python/cpython/blob/v3.11.9/Lib/importlib/_bootstrap_external.py#L1585
            # https://github.com/python/cpython/blob/v3.11.9/Lib/importlib/_bootstrap_external.py#L622-L639
            #
            loader, portions = self.find_loader(fullname)
            return loader


# Helper for enforcing module name in PyiFrozenLoader methods.
def _check_name(method):
    def _check_name_wrapper(self, name, *args, **kwargs):
        if self.name != name:
            raise ImportError(f'loader for {self.name} cannot handle {name}', name=name)
        return method(self, name, *args, **kwargs)

    return _check_name_wrapper


class PyiFrozenLoader:
    """
    PyInstaller's frozen loader for modules in the PYZ archive, which are discovered by PyiFrozenFinder.

    Since this loader is instantiated only from PyiFrozenFinder and since each loader instance is tied to a specific
    module, the fact that the loader was instantiated serves as the proof that the module exists in the PYZ archive.
    Hence, we can avoid any additional validation in the implementation of the loader's methods.
    """
    def __init__(self, name, pyz_archive, pyz_entry_name, is_package):
        # Store the reference to PYZ archive (for code object retrieval), as well as full PYZ entry name
        # and typecode, all of which are passed from the PyiFrozenFinder.
        self._pyz_archive = pyz_archive
        self._pyz_entry_name = pyz_entry_name
        self._is_package = is_package

        # Compute the module file path, as if module was located on filesystem.
        #
        # Rather than returning path to the .pyc file, return the path to .py file - which might actually exist, if it
        # was explicitly collected into the frozen application). This improves compliance with
        # https://docs.python.org/3/library/importlib.html#importlib.abc.ExecutionLoader.get_filename
        # as well as general compatibility with 3rd party code that blindly assumes that module's file path points to
        # the source .py file.
        #
        # NOTE: since we are using sys._MEIPASS as prefix, we need to construct path from full PYZ entry name
        # (so that a module with `name`=`jaraco.text` and `pyz_entry_name`=`setuptools._vendor.jaraco.text`
        # ends up with path set to `sys._MEIPASS/setuptools/_vendor/jaraco/text/__init__.pyc` instead of
        # `sys._MEIPASS/jaraco/text/__init__.pyc`).
        if is_package:
            module_file = os.path.join(sys._MEIPASS, pyz_entry_name.replace('.', os.path.sep), '__init__.py')
        else:
            module_file = os.path.join(sys._MEIPASS, pyz_entry_name.replace('.', os.path.sep) + '.py')

        # These properties are defined as part of importlib.abc.FileLoader. They are used by our implementation
        # (e.g., module name validation, get_filename(), get_source(), get_resource_reader()), and might also be used
        # by 3rd party code that naively expects to be dealing with a FileLoader instance.
        self.name = name  # The name of the module the loader can handle.
        self.path = module_file  # Path to the file of the module

    #-- Core PEP451 loader functionality as defined by importlib.abc.Loader
    # https://docs.python.org/3/library/importlib.html#importlib.abc.Loader
    def create_module(self, spec):
        """
        A method that returns the module object to use when importing a module. This method may return None, indicating
        that default module creation semantics should take place.

        https://docs.python.org/3/library/importlib.html#importlib.abc.Loader.create_module
        """
        return None

    def exec_module(self, module):
        """
        A method that executes the module in its own namespace when a module is imported or reloaded. The module
        should already be initialized when exec_module() is called. When this method exists, create_module()
        must be defined.

        https://docs.python.org/3/library/importlib.html#importlib.abc.Loader.exec_module
        """
        spec = module.__spec__
        bytecode = self.get_code(spec.name)  # NOTE: get_code verifies that `spec.name` matches `self.name`!
        if bytecode is None:
            raise RuntimeError(f"Failed to retrieve bytecode for {spec.name!r}!")

        # Set by the import machinery
        assert hasattr(module, '__file__')

        # If `submodule_search_locations` is not None, this is a package; set __path__.
        if spec.submodule_search_locations is not None:
            module.__path__ = spec.submodule_search_locations

        exec(bytecode, module.__dict__)

    # The following method is part of legacy PEP302 loader interface. It has been deprecated since python 3.4, and
    # slated for removal in python 3.12, although that has not happened yet. Provide compatibility shim to accommodate
    # code that might still be using it.
    if True:

        @_check_name
        def load_module(self, fullname):
            """
            A legacy method for loading a module. If the module cannot be loaded, ImportError is raised, otherwise the
            loaded module is returned.

            Deprecated since python 3.4, slated for removal in 3.12 (but still present in python's own FileLoader in
            both v3.12.4 and v3.13.0rc1).
            """
            # Based on:
            # https://github.com/python/cpython/blob/v3.11.9/Lib/importlib/_bootstrap_external.py#L942-L945
            import importlib._bootstrap as _bootstrap
            return _bootstrap._load_module_shim(self, fullname)

    #-- PEP302 protocol extensions as defined by importlib.abc.ExecutionLoader
    # https://docs.python.org/3/library/importlib.html#importlib.abc.ExecutionLoader
    @_check_name
    def get_filename(self, fullname):
        """
        A method that is to return the value of __file__ for the specified module. If no path is available, ImportError
        is raised.

        If source code is available, then the method should return the path to the source file, regardless of whether a
        bytecode was used to load the module.

        https://docs.python.org/3/library/importlib.html#importlib.abc.ExecutionLoader.get_filename
        """
        return self.path

    #-- PEP302 protocol extensions as defined by importlib.abc.InspectLoader
    # https://docs.python.org/3/library/importlib.html#importlib.abc.InspectLoader
    @_check_name
    def get_code(self, fullname):
        """
        Return the code object for a module, or None if the module does not have a code object (as would be the case,
        for example, for a built-in module). Raise an ImportError if loader cannot find the requested module.

        https://docs.python.org/3/library/importlib.html#importlib.abc.InspectLoader.get_code
        """
        return self._pyz_archive.extract(self._pyz_entry_name)

    @_check_name
    def get_source(self, fullname):
        """
        A method to return the source of a module. It is returned as a text string using universal newlines, translating
        all recognized line separators into '\n' characters. Returns None if no source is available (e.g. a built-in
        module). Raises ImportError if the loader cannot find the module specified.

        https://docs.python.org/3/library/importlib.html#importlib.abc.InspectLoader.get_source
        """
        # The `path` attribute (which is also returned from `get_filename()`) already points to where the source .py
        # file should exist, if it is available.
        filename = self.path

        try:
            # Read in binary mode, then decode
            with open(filename, 'rb') as fp:
                source_bytes = fp.read()
            return _decode_source(source_bytes)
        except FileNotFoundError:
            pass

        # Source code is unavailable.
        return None

    @_check_name
    def is_package(self, fullname):
        """
        A method to return a true value if the module is a package, a false value otherwise. ImportError is raised if
        the loader cannot find the module.

        https://docs.python.org/3/library/importlib.html#importlib.abc.InspectLoader.is_package
        """
        return self._is_package

    #-- PEP302 protocol extensions as dfined by importlib.abc.ResourceLoader
    # https://docs.python.org/3/library/importlib.html#importlib.abc.ResourceLoader
    def get_data(self, path):
        """
        A method to return the bytes for the data located at path. Loaders that have a file-like storage back-end that
        allows storing arbitrary data can implement this abstract method to give direct access to the data stored.
        OSError is to be raised if the path cannot be found. The path is expected to be constructed using a module’s
        __file__ attribute or an item from a package’s __path__.

        https://docs.python.org/3/library/importlib.html#importlib.abc.ResourceLoader.get_data
        """
        # Try to fetch the data from the filesystem. Since __file__ attribute works properly, just try to open the file
        # and read it.
        with open(path, 'rb') as fp:
            return fp.read()

    #-- Support for `importlib.resources`.
    @_check_name
    def get_resource_reader(self, fullname):
        """
        Return resource reader compatible with `importlib.resources`.
        """
        return PyiFrozenResourceReader(self)


class PyiFrozenResourceReader:
    """
    Resource reader for importlib.resources / importlib_resources support.

    Supports only on-disk resources, which should cover the typical use cases, i.e., the access to data files;
    PyInstaller collects data files onto filesystem, and as of v6.0.0, the embedded PYZ archive is guaranteed
    to contain only .pyc modules.

    When listing resources, source .py files will not be listed as they are not collected by default. Similarly,
    sub-directories that contained only .py files are not reconstructed on filesystem, so they will not be listed,
    either. If access to .py files is required for whatever reason, they need to be explicitly collected as data files
    anyway, which will place them on filesystem and make them appear as resources.

    For on-disk resources, we *must* return path compatible with pathlib.Path() in order to avoid copy to a temporary
    file, which might break under some circumstances, e.g., metpy with importlib_resources back-port, due to:
    https://github.com/Unidata/MetPy/blob/a3424de66a44bf3a92b0dcacf4dff82ad7b86712/src/metpy/plots/wx_symbols.py#L24-L25
    (importlib_resources tries to use 'fonts/wx_symbols.ttf' as a temporary filename suffix, which fails as it contains
    a separator).

    Furthermore, some packages expect files() to return either pathlib.Path or zipfile.Path, e.g.,
    https://github.com/tensorflow/datasets/blob/master/tensorflow_datasets/core/utils/resource_utils.py#L81-L97
    This makes implementation of mixed support for on-disk and embedded resources using importlib.abc.Traversable
    protocol rather difficult.

    So in order to maximize compatibility with unfrozen behavior, the below implementation is basically equivalent of
    importlib.readers.FileReader from python 3.10:
      https://github.com/python/cpython/blob/839d7893943782ee803536a47f1d4de160314f85/Lib/importlib/readers.py#L11
    and its underlying classes, importlib.abc.TraversableResources and importlib.abc.ResourceReader:
      https://github.com/python/cpython/blob/839d7893943782ee803536a47f1d4de160314f85/Lib/importlib/abc.py#L422
      https://github.com/python/cpython/blob/839d7893943782ee803536a47f1d4de160314f85/Lib/importlib/abc.py#L312
    """
    def __init__(self, loader):
        # Local import to avoid including `pathlib` and its dependencies in `base_library.zip`
        import pathlib
        # This covers both modules and (regular) packages. Note that PEP-420 namespace packages are not handled by this
        # resource reader (since they are not handled by PyiFrozenLoader, which uses this reader).
        self.path = pathlib.Path(loader.path).parent

    def open_resource(self, resource):
        return self.files().joinpath(resource).open('rb')

    def resource_path(self, resource):
        return str(self.path.joinpath(resource))

    def is_resource(self, path):
        return self.files().joinpath(path).is_file()

    def contents(self):
        return (item.name for item in self.files().iterdir())

    def files(self):
        return self.path


class PyiFrozenEntryPointLoader:
    """
    A special loader that enables retrieval of the code-object for the __main__ module.
    """
    def __repr__(self):
        return self.__class__.__name__

    def get_code(self, fullname):
        if fullname == '__main__':
            # Special handling for __main__ module; the bootloader should store code object to _pyi_main_co
            # attribute of the module.
            return sys.modules['__main__']._pyi_main_co

        raise ImportError(f'{self} cannot handle module {fullname!r}')


def install():
    """
    Install PyInstaller's frozen finders/loaders/importers into python's import machinery.
    """
    # Setup PYZ archive reader.
    #
    # The bootloader should store the path to PYZ archive (the path to the PKG archive and the offset within it; for
    # executable-embedded archive, this is for example /path/executable_name?117568) into _pyinstaller_pyz
    # attribute of the sys module.
    global pyz_archive

    if not hasattr(sys, '_pyinstaller_pyz'):
        raise RuntimeError("Bootloader did not set sys._pyinstaller_pyz!")

    try:
        pyz_archive = pyimod01_archive.ZlibArchiveReader(sys._pyinstaller_pyz, check_pymagic=True)
    except Exception as e:
        raise RuntimeError("Failed to setup PYZ archive reader!") from e

    delattr(sys, '_pyinstaller_pyz')

    # On Windows, there is finder called `_frozen_importlib.WindowsRegistryFinder`, which looks for Python module
    # locations in Windows registry. The frozen application should not look for those, so remove this finder
    # from `sys.meta_path`.
    for entry in sys.meta_path:
        if getattr(entry, '__name__', None) == 'WindowsRegistryFinder':
            sys.meta_path.remove(entry)
            break

    # Insert our hook for `PyiFrozenFinder` into `sys.path_hooks`. Place it after `zipimporter`, if available.
    for idx, entry in enumerate(sys.path_hooks):
        if getattr(entry, '__name__', None) == 'zipimporter':
            trace(f"PyInstaller: inserting our finder hook at index {idx + 1} in sys.path_hooks.")
            sys.path_hooks.insert(idx + 1, PyiFrozenFinder.path_hook)
            break
    else:
        trace("PyInstaller: zipimporter hook not found in sys.path_hooks! Prepending our finder hook to the list.")
        sys.path_hooks.insert(0, PyiFrozenFinder.path_hook)

    # Monkey-patch `zipimporter.get_source` to allow loading out-of-zip source .py files for modules that are
    # in `base_library.zip`.
    _patch_zipimporter_get_source()

    # Python might have already created a `FileFinder` for `sys._MEIPASS`. Remove the entry from path importer cache,
    # so that next loading attempt creates `PyiFrozenFinder` instead. This could probably be avoided altogether if
    # we refrained from adding `sys._MEIPASS` to `sys.path` until our importer hooks is in place.
    sys.path_importer_cache.pop(sys._MEIPASS, None)

    # Set the PyiFrozenEntryPointLoader as loader for __main__, in order for python to treat __main__ as a module
    # instead of a built-in, and to allow its code object to be retrieved.
    try:
        sys.modules['__main__'].__loader__ = PyiFrozenEntryPointLoader()
    except Exception:
        pass

    # Apply hack for python >= 3.11 and its frozen stdlib modules.
    if sys.version_info >= (3, 11):
        _fixup_frozen_stdlib()


# A hack for python >= 3.11 and its frozen stdlib modules. Unless `sys._stdlib_dir` is set, these modules end up
# missing __file__ attribute, which causes problems with 3rd party code. At the time of writing, python interpreter
# configuration API does not allow us to influence `sys._stdlib_dir` - it always resets it to `None`. Therefore,
# we manually set the path, and fix __file__ attribute on modules.
def _fixup_frozen_stdlib():
    import _imp  # built-in

    # If sys._stdlib_dir is None or empty, override it with sys._MEIPASS
    if not sys._stdlib_dir:
        try:
            sys._stdlib_dir = sys._MEIPASS
        except AttributeError:
            pass

    # The sys._stdlib_dir set above should affect newly-imported python-frozen modules. However, most of them have
    # been already imported during python initialization and our bootstrap, so we need to retroactively fix their
    # __file__ attribute.
    for module_name, module in sys.modules.items():
        if not _imp.is_frozen(module_name):
            continue

        is_pkg = _imp.is_frozen_package(module_name)

        # Determine "real" name from __spec__.loader_state.
        loader_state = module.__spec__.loader_state

        orig_name = loader_state.origname
        if is_pkg:
            orig_name += '.__init__'

        # We set suffix to .pyc to be consistent with our PyiFrozenLoader.
        filename = os.path.join(sys._MEIPASS, *orig_name.split('.')) + '.pyc'

        # Fixup the __file__ attribute
        if not hasattr(module, '__file__'):
            try:
                module.__file__ = filename
            except AttributeError:
                pass

        # Fixup the loader_state.filename
        # Except for _frozen_importlib (importlib._bootstrap), whose loader_state.filename appears to be left at
        # None in python.
        if loader_state.filename is None and orig_name != 'importlib._bootstrap':
            loader_state.filename = filename


# Monkey-patch the `get_source` implementation of python's `zipimport.zipimporter` with our custom implementation that
# looks up for source files in top-level application directory instead of within the zip file. This allows us to collect
# source .py files for modules that are collected in the `base_library.zip` in the same way as for modules in the PYZ
# archive.
def _patch_zipimporter_get_source():
    import zipimport

    _orig_get_source = zipimport.zipimporter.get_source

    def _get_source(self, fullname):
        # Call original implementation first, in case we are dealing with a zip file other than `base_library.zip` (or
        # if the source .py file is actually in there, for whatever reason). This also implicitly validates the module
        # name, as it raises exception if module does not exist and returns None if module exists but the source code
        # is not present in the archive.
        source = _orig_get_source(self, fullname)
        if source is not None:
            return source

        # Our override should apply only to `base_library.zip`.
        if os.path.basename(self.archive) != 'base_library.zip':
            return None

        # Translate module/package name into .py filename in the top-level application directory.
        if self.is_package(fullname):
            filename = os.path.join(*fullname.split('.'), '__init__.py')
        else:
            filename = os.path.join(*fullname.split('.')) + '.py'
        filename = os.path.join(_RESOLVED_TOP_LEVEL_DIRECTORY, filename)

        try:
            # Read in binary mode, then decode
            with open(filename, 'rb') as fp:
                source_bytes = fp.read()
            return _decode_source(source_bytes)
        except FileNotFoundError:
            pass

        # Source code is unavailable.
        return None

    zipimport.zipimporter.get_source = _get_source



# NOTE: if you edit this file, be careful not to change the encoding!
#
# The comment line below intentionally contains a non-ASCII character in local encoding and fails to declare the
# encoding using PEP263 encoding header.
#
# problematic character: é (e-acute in ISO-8859-15)
#
# While python is able to load and run the module, retrieving its source code via loader's `get_source`
# method is expected to raise an error:
#
# ```
# import mymodule1
# mymodule1.__loader__.get_source('mymodule1')
# ```
#
# ```
# Traceback (most recent call last):
#   File "<stdin>", line 1, in <module>
#   File "<frozen importlib._bootstrap_external>", line 997, in get_source
#   File "<frozen importlib._bootstrap_external>", line 770, in decode_source
# UnicodeDecodeError: 'utf-8' codec can't decode byte 0xe9 in position 298: invalid continuation byte
# ```

def hello():
    return "hello"



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, sys, errno, re, shutil, stat
try:
    import cPickle
except ImportError:
    import pickle as cPickle
from waflib import Node, Runner, TaskGen, Utils, ConfigSet, Task, Logs, Options, Context, Errors

CACHE_DIR = 'c4che'
CACHE_SUFFIX = '_cache.py'
INSTALL = 1337
UNINSTALL = -1337
SAVED_ATTRS = 'root node_sigs task_sigs imp_sigs raw_deps node_deps'.split()
CFG_FILES = 'cfg_files'
POST_AT_ONCE = 0
POST_LAZY = 1
PROTOCOL = -1
if sys.platform == 'cli':
    PROTOCOL = 0


class BuildContext(Context.Context):
    '''executes the build'''
    cmd = 'build'
    variant = ''

    def __init__(self, **kw):
        super(BuildContext, self).__init__(**kw)
        self.is_install = 0
        self.top_dir = kw.get('top_dir', Context.top_dir)
        self.out_dir = kw.get('out_dir', Context.out_dir)
        self.run_dir = kw.get('run_dir', Context.run_dir)
        self.launch_dir = Context.launch_dir
        self.post_mode = POST_LAZY
        self.cache_dir = kw.get('cache_dir')
        if not self.cache_dir:
            self.cache_dir = os.path.join(self.out_dir, CACHE_DIR)
        self.all_envs = {}
        self.node_sigs = {}
        self.task_sigs = {}
        self.imp_sigs = {}
        self.node_deps = {}
        self.raw_deps = {}
        self.task_gen_cache_names = {}
        self.jobs = Options.options.jobs
        self.targets = Options.options.targets
        self.keep = Options.options.keep
        self.progress_bar = Options.options.progress_bar
        self.deps_man = Utils.defaultdict(list)
        self.current_group = 0
        self.groups = []
        self.group_names = {}
        for v in SAVED_ATTRS:
            if not hasattr(self, v):
                setattr(self, v, {})

    def get_variant_dir(self):
        if not self.variant:
            return self.out_dir
        return os.path.join(self.out_dir, os.path.normpath(self.variant))

    variant_dir = property(get_variant_dir, None)

    def __call__(self, *k, **kw):
        kw['bld'] = self
        ret = TaskGen.task_gen(*k, **kw)
        self.task_gen_cache_names = {}
        self.add_to_group(ret, group=kw.get('group'))
        return ret

    def __copy__(self):
        raise Errors.WafError('build contexts cannot be copied')

    def load_envs(self):
        node = self.root.find_node(self.cache_dir)
        if not node:
            raise Errors.WafError('The project was not configured: run "waf configure" first!')
        lst = node.ant_glob('**/*%s' % CACHE_SUFFIX, quiet=True)
        if not lst:
            raise Errors.WafError('The cache directory is empty: reconfigure the project')
        for x in lst:
            name = x.path_from(node).replace(CACHE_SUFFIX, '').replace('\\', '/')
            env = ConfigSet.ConfigSet(x.abspath())
            self.all_envs[name] = env
            for f in env[CFG_FILES]:
                newnode = self.root.find_resource(f)
                if not newnode or not newnode.exists():
                    raise Errors.WafError('Missing configuration file %r, reconfigure the project!' % f)

    def init_dirs(self):
        if not (os.path.isabs(self.top_dir) and os.path.isabs(self.out_dir)):
            raise Errors.WafError('The project was not configured: run "waf configure" first!')
        self.path = self.srcnode = self.root.find_dir(self.top_dir)
        self.bldnode = self.root.make_node(self.variant_dir)
        self.bldnode.mkdir()

    def execute(self):
        self.restore()
        if not self.all_envs:
            self.load_envs()
        self.execute_build()

    def execute_build(self):
        Logs.info("Waf: Entering directory `%s'", self.variant_dir)
        self.recurse([self.run_dir])
        self.pre_build()
        self.timer = Utils.Timer()
        try:
            self.compile()
        finally:
            if self.progress_bar == 1 and sys.stderr.isatty():
                c = self.producer.processed or 1
                m = self.progress_line(c, c, Logs.colors.BLUE, Logs.colors.NORMAL)
                Logs.info(m, extra={'stream': sys.stderr, 'c1': Logs.colors.cursor_off, 'c2': Logs.colors.cursor_on})
            Logs.info("Waf: Leaving directory `%s'", self.variant_dir)
        try:
            self.producer.bld = None
            del self.producer
        except AttributeError:
            pass
        self.post_build()

    def restore(self):
        try:
            env = ConfigSet.ConfigSet(os.path.join(self.cache_dir, 'build.config.py'))
        except EnvironmentError:
            pass
        else:
            if env.version < Context.HEXVERSION:
                raise Errors.WafError('Project was configured with a different version of Waf, please reconfigure it')
            for t in env.tools:
                self.setup(**t)
        dbfn = os.path.join(self.variant_dir, Context.DBFILE)
        try:
            data = Utils.readf(dbfn, 'rb')
        except (EnvironmentError, EOFError):
            Logs.debug('build: Could not load the build cache %s (missing)', dbfn)
        else:
            try:
                Node.pickle_lock.acquire()
                Node.Nod3 = self.node_class
                try:
                    data = cPickle.loads(data)
                except Exception as e:
                    Logs.debug('build: Could not pickle the build cache %s: %r', dbfn, e)
                else:
                    for x in SAVED_ATTRS:
                        setattr(self, x, data.get(x, {}))
            finally:
                Node.pickle_lock.release()
        self.init_dirs()

    def store(self):
        data = {}
        for x in SAVED_ATTRS:
            data[x] = getattr(self, x)
        db = os.path.join(self.variant_dir, Context.DBFILE)
        try:
            Node.pickle_lock.acquire()
            Node.Nod3 = self.node_class
            x = cPickle.dumps(data, PROTOCOL)
        finally:
            Node.pickle_lock.release()
        Utils.writef(db + '.tmp', x, m='wb')
        try:
            st = os.stat(db)
            os.remove(db)
            if not Utils.is_win32:
                os.chown(db + '.tmp', st.st_uid, st.st_gid)
        except (AttributeError, OSError):
            pass
        os.rename(db + '.tmp', db)

    def compile(self):
        Logs.debug('build: compile()')
        self.producer = Runner.Parallel(self, self.jobs)
        self.producer.biter = self.get_build_iterator()
        try:
            self.producer.start()
        except KeyboardInterrupt:
            if self.is_dirty():
                self.store()
            raise
        else:
            if self.is_dirty():
                self.store()
        if self.producer.error:
            raise Errors.BuildError(self.producer.error)

    def is_dirty(self):
        return self.producer.dirty

    def setup(self, tool, tooldir=None, funs=None):
        if isinstance(tool, list):
            for i in tool:
                self.setup(i, tooldir)
            return
        module = Context.load_tool(tool, tooldir)
        if hasattr(module, "setup"):
            module.setup(self)

    def get_env(self):
        try:
            return self.all_envs[self.variant]
        except KeyError:
            return self.all_envs['']

    def set_env(self, val):
        self.all_envs[self.variant] = val

    env = property(get_env, set_env)

    def add_manual_dependency(self, path, value):
        if not path:
            raise ValueError('Invalid input path %r' % path)
        if isinstance(path, Node.Node):
            node = path
        elif os.path.isabs(path):
            node = self.root.find_resource(path)
        else:
            node = self.path.find_resource(path)
        if not node:
            raise ValueError('Could not find the path %r' % path)
        if isinstance(value, list):
            self.deps_man[node].extend(value)
        else:
            self.deps_man[node].append(value)

    def launch_node(self):
        try:
            return self.p_ln
        except AttributeError:
            self.p_ln = self.root.find_dir(self.launch_dir)
            return self.p_ln

    def hash_env_vars(self, env, vars_lst):
        if not env.table:
            env = env.parent
            if not env:
                return Utils.SIG_NIL
        idx = str(id(env)) + str(vars_lst)
        try:
            cache = self.cache_env
        except AttributeError:
            cache = self.cache_env = {}
        else:
            try:
                return self.cache_env[idx]
            except KeyError:
                pass
        lst = [env[a] for a in vars_lst]
        cache[idx] = ret = Utils.h_list(lst)
        Logs.debug('envhash: %s %r', Utils.to_hex(ret), lst)
        return ret

    def get_tgen_by_name(self, name):
        cache = self.task_gen_cache_names
        if not cache:
            for g in self.groups:
                for tg in g:
                    try:
                        cache[tg.name] = tg
                    except AttributeError:
                        pass
        try:
            return cache[name]
        except KeyError:
            raise Errors.WafError('Could not find a task generator for the name %r' % name)

    def progress_line(self, idx, total, col1, col2):
        if not sys.stderr.isatty():
            return ''
        n = len(str(total))
        Utils.rot_idx += 1
        ind = Utils.rot_chr[Utils.rot_idx % 4]
        pc = (100. * idx) / total
        fs = "[%%%dd/%%d][%%s%%2d%%%%%%s][%s][" % (n, ind)
        left = fs % (idx, total, col1, pc, col2)
        right = '][%s%s%s]' % (col1, self.timer, col2)
        cols = Logs.get_term_cols() - len(left) - len(right) + 2 * len(col1) + 2 * len(col2)
        if cols < 7:
            cols = 7
        ratio = ((cols * idx) // total) - 1
        bar = ('=' * ratio + '>').ljust(cols)
        msg = Logs.indicator % (left, bar, right)
        return msg

    def declare_chain(self, *k, **kw):
        return TaskGen.declare_chain(*k, **kw)

    def pre_build(self):
        for m in getattr(self, 'pre_funs', []):
            m(self)

    def post_build(self):
        for m in getattr(self, 'post_funs', []):
            m(self)

    def add_pre_fun(self, meth):
        try:
            self.pre_funs.append(meth)
        except AttributeError:
            self.pre_funs = [meth]

    def add_post_fun(self, meth):
        try:
            self.post_funs.append(meth)
        except AttributeError:
            self.post_funs = [meth]

    def get_group(self, x):
        if not self.groups:
            self.add_group()
        if x is None:
            return self.groups[self.current_group]
        if x in self.group_names:
            return self.group_names[x]
        return self.groups[x]

    def add_to_group(self, tgen, group=None):
        assert (isinstance(tgen, TaskGen.task_gen) or isinstance(tgen, Task.Task))
        tgen.bld = self
        self.get_group(group).append(tgen)

    def get_group_name(self, g):
        if not isinstance(g, list):
            g = self.groups[g]
        for x in self.group_names:
            if id(self.group_names[x]) == id(g):
                return x
        return ''

    def get_group_idx(self, tg):
        se = id(tg)
        for i, tmp in enumerate(self.groups):
            for t in tmp:
                if id(t) == se:
                    return i
        return None

    def add_group(self, name=None, move=True):
        if name and name in self.group_names:
            raise Errors.WafError('add_group: name %s already present', name)
        g = []
        self.group_names[name] = g
        self.groups.append(g)
        if move:
            self.current_group = len(self.groups) - 1

    def set_group(self, idx):
        if isinstance(idx, str):
            g = self.group_names[idx]
            for i, tmp in enumerate(self.groups):
                if id(g) == id(tmp):
                    self.current_group = i
                    break
        else:
            self.current_group = idx

    def total(self):
        total = 0
        for group in self.groups:
            for tg in group:
                try:
                    total += len(tg.tasks)
                except AttributeError:
                    total += 1
        return total

    def get_targets(self):
        to_post = []
        min_grp = 0
        for name in self.targets.split(','):
            tg = self.get_tgen_by_name(name)
            m = self.get_group_idx(tg)
            if m > min_grp:
                min_grp = m
                to_post = [tg]
            elif m == min_grp:
                to_post.append(tg)
        return (min_grp, to_post)

    def get_all_task_gen(self):
        lst = []
        for g in self.groups:
            lst.extend(g)
        return lst

    def post_group(self):
        def tgpost(tg):
            try:
                f = tg.post
            except AttributeError:
                pass
            else:
                f()

        if self.targets == '*':
            for tg in self.groups[self.current_group]:
                tgpost(tg)
        elif self.targets:
            if self.current_group < self._min_grp:
                for tg in self.groups[self.current_group]:
                    tgpost(tg)
            else:
                for tg in self._exact_tg:
                    tg.post()
        else:
            ln = self.launch_node()
            if ln.is_child_of(self.bldnode):
                Logs.warn('Building from the build directory, forcing --targets=*')
                ln = self.srcnode
            elif not ln.is_child_of(self.srcnode):
                Logs.warn(
                    'CWD %s is not under %s, forcing --targets=* (run distclean?)', ln.abspath(), self.srcnode.abspath()
                )
                ln = self.srcnode

            def is_post(tg, ln):
                try:
                    p = tg.path
                except AttributeError:
                    pass
                else:
                    if p.is_child_of(ln):
                        return True

            def is_post_group():
                for i, g in enumerate(self.groups):
                    if i > self.current_group:
                        for tg in g:
                            if is_post(tg, ln):
                                return True

            if self.post_mode == POST_LAZY and ln != self.srcnode:
                if is_post_group():
                    ln = self.srcnode
            for tg in self.groups[self.current_group]:
                if is_post(tg, ln):
                    tgpost(tg)

    def get_tasks_group(self, idx):
        tasks = []
        for tg in self.groups[idx]:
            try:
                tasks.extend(tg.tasks)
            except AttributeError:
                tasks.append(tg)
        return tasks

    def get_build_iterator(self):
        if self.targets and self.targets != '*':
            (self._min_grp, self._exact_tg) = self.get_targets()
        if self.post_mode != POST_LAZY:
            for self.current_group, _ in enumerate(self.groups):
                self.post_group()
        for self.current_group, _ in enumerate(self.groups):
            if self.post_mode != POST_AT_ONCE:
                self.post_group()
            tasks = self.get_tasks_group(self.current_group)
            Task.set_file_constraints(tasks)
            Task.set_precedence_constraints(tasks)
            self.cur_tasks = tasks
            if tasks:
                yield tasks
        while 1:
            yield []

    def install_files(self, dest, files, **kw):
        assert (dest)
        tg = self(features='install_task', install_to=dest, install_from=files, **kw)
        tg.dest = tg.install_to
        tg.type = 'install_files'
        if not kw.get('postpone', True):
            tg.post()
        return tg

    def install_as(self, dest, srcfile, **kw):
        assert (dest)
        tg = self(features='install_task', install_to=dest, install_from=srcfile, **kw)
        tg.dest = tg.install_to
        tg.type = 'install_as'
        if not kw.get('postpone', True):
            tg.post()
        return tg

    def symlink_as(self, dest, src, **kw):
        assert (dest)
        tg = self(features='install_task', install_to=dest, install_from=src, **kw)
        tg.dest = tg.install_to
        tg.type = 'symlink_as'
        tg.link = src
        if not kw.get('postpone', True):
            tg.post()
        return tg


@TaskGen.feature('install_task')
@TaskGen.before_method('process_rule', 'process_source')
def process_install_task(self):
    self.add_install_task(**self.__dict__)


@TaskGen.taskgen_method
def add_install_task(self, **kw):
    if not self.bld.is_install:
        return
    if not kw['install_to']:
        return
    if kw['type'] == 'symlink_as' and Utils.is_win32:
        if kw.get('win32_install'):
            kw['type'] = 'install_as'
        else:
            return
    tsk = self.install_task = self.create_task('inst')
    tsk.chmod = kw.get('chmod', Utils.O644)
    tsk.link = kw.get('link', '') or kw.get('install_from', '')
    tsk.relative_trick = kw.get('relative_trick', False)
    tsk.type = kw['type']
    tsk.install_to = tsk.dest = kw['install_to']
    tsk.install_from = kw['install_from']
    tsk.relative_base = kw.get('cwd') or kw.get('relative_base', self.path)
    tsk.install_user = kw.get('install_user')
    tsk.install_group = kw.get('install_group')
    tsk.init_files()
    if not kw.get('postpone', True):
        tsk.run_now()
    return tsk


@TaskGen.taskgen_method
def add_install_files(self, **kw):
    kw['type'] = 'install_files'
    return self.add_install_task(**kw)


@TaskGen.taskgen_method
def add_install_as(self, **kw):
    kw['type'] = 'install_as'
    return self.add_install_task(**kw)


@TaskGen.taskgen_method
def add_symlink_as(self, **kw):
    kw['type'] = 'symlink_as'
    return self.add_install_task(**kw)


class inst(Task.Task):
    def __str__(self):
        return ''

    def uid(self):
        lst = self.inputs + self.outputs + [self.link, self.generator.path.abspath()]
        return Utils.h_list(lst)

    def init_files(self):
        if self.type == 'symlink_as':
            inputs = []
        else:
            inputs = self.generator.to_nodes(self.install_from)
            if self.type == 'install_as':
                assert len(inputs) == 1
        self.set_inputs(inputs)
        dest = self.get_install_path()
        outputs = []
        if self.type == 'symlink_as':
            if self.relative_trick:
                self.link = os.path.relpath(self.link, os.path.dirname(dest))
            outputs.append(self.generator.bld.root.make_node(dest))
        elif self.type == 'install_as':
            outputs.append(self.generator.bld.root.make_node(dest))
        else:
            for y in inputs:
                if self.relative_trick:
                    destfile = os.path.join(dest, y.path_from(self.relative_base))
                else:
                    destfile = os.path.join(dest, y.name)
                outputs.append(self.generator.bld.root.make_node(destfile))
        self.set_outputs(outputs)

    def runnable_status(self):
        ret = super(inst, self).runnable_status()
        if ret == Task.SKIP_ME and self.generator.bld.is_install:
            return Task.RUN_ME
        return ret

    def post_run(self):
        pass

    def get_install_path(self, destdir=True):
        if isinstance(self.install_to, Node.Node):
            dest = self.install_to.abspath()
        else:
            dest = os.path.normpath(Utils.subst_vars(self.install_to, self.env))
        if not os.path.isabs(dest):
            dest = os.path.join(self.env.PREFIX, dest)
        if destdir and Options.options.destdir:
            dest = os.path.join(Options.options.destdir, os.path.splitdrive(dest)[1].lstrip(os.sep))
        return dest

    def copy_fun(self, src, tgt):
        if Utils.is_win32 and len(tgt) > 259 and not tgt.startswith('\\\\?\\'):
            tgt = '\\\\?\\' + tgt
        shutil.copy2(src, tgt)
        self.fix_perms(tgt)

    def rm_empty_dirs(self, tgt):
        while tgt:
            tgt = os.path.dirname(tgt)
            try:
                os.rmdir(tgt)
            except OSError:
                break

    def run(self):
        is_install = self.generator.bld.is_install
        if not is_install:
            return
        for x in self.outputs:
            if is_install == INSTALL:
                x.parent.mkdir()
        if self.type == 'symlink_as':
            fun = is_install == INSTALL and self.do_link or self.do_unlink
            fun(self.link, self.outputs[0].abspath())
        else:
            fun = is_install == INSTALL and self.do_install or self.do_uninstall
            launch_node = self.generator.bld.launch_node()
            for x, y in zip(self.inputs, self.outputs):
                fun(x.abspath(), y.abspath(), x.path_from(launch_node))

    def run_now(self):
        status = self.runnable_status()
        if status not in (Task.RUN_ME, Task.SKIP_ME):
            raise Errors.TaskNotReady('Could not process %r: status %r' % (self, status))
        self.run()
        self.hasrun = Task.SUCCESS

    def do_install(self, src, tgt, lbl, **kw):
        if not Options.options.force:
            try:
                st1 = os.stat(tgt)
                st2 = os.stat(src)
            except OSError:
                pass
            else:
                if st1.st_mtime + 2 >= st2.st_mtime and st1.st_size == st2.st_size:
                    if not self.generator.bld.progress_bar:
                        c1 = Logs.colors.NORMAL
                        c2 = Logs.colors.BLUE
                        Logs.info('%s- install %s%s%s (from %s)', c1, c2, tgt, c1, lbl)
                    return False
        if not self.generator.bld.progress_bar:
            c1 = Logs.colors.NORMAL
            c2 = Logs.colors.BLUE
            Logs.info('%s+ install %s%s%s (from %s)', c1, c2, tgt, c1, lbl)
        try:
            os.chmod(tgt, Utils.O644 | stat.S_IMODE(os.stat(tgt).st_mode))
        except EnvironmentError:
            pass
        try:
            os.remove(tgt)
        except OSError:
            pass
        try:
            self.copy_fun(src, tgt)
        except EnvironmentError as e:
            if not os.path.exists(src):
                Logs.error('File %r does not exist', src)
            elif not os.path.isfile(src):
                Logs.error('Input %r is not a file', src)
            raise Errors.WafError('Could not install the file %r' % tgt, e)

    def fix_perms(self, tgt):
        if not Utils.is_win32:
            user = getattr(self, 'install_user', None) or getattr(self.generator, 'install_user', None)
            group = getattr(self, 'install_group', None) or getattr(self.generator, 'install_group', None)
            if user or group:
                Utils.lchown(tgt, user or -1, group or -1)
        if not os.path.islink(tgt):
            os.chmod(tgt, self.chmod)

    def do_link(self, src, tgt, **kw):
        if os.path.islink(tgt) and os.readlink(tgt) == src:
            if not self.generator.bld.progress_bar:
                c1 = Logs.colors.NORMAL
                c2 = Logs.colors.BLUE
                Logs.info('%s- symlink %s%s%s (to %s)', c1, c2, tgt, c1, src)
        else:
            try:
                os.remove(tgt)
            except OSError:
                pass
            if not self.generator.bld.progress_bar:
                c1 = Logs.colors.NORMAL
                c2 = Logs.colors.BLUE
                Logs.info('%s+ symlink %s%s%s (to %s)', c1, c2, tgt, c1, src)
            os.symlink(src, tgt)
            self.fix_perms(tgt)

    def do_uninstall(self, src, tgt, lbl, **kw):
        if not self.generator.bld.progress_bar:
            c1 = Logs.colors.NORMAL
            c2 = Logs.colors.BLUE
            Logs.info('%s- remove %s%s%s', c1, c2, tgt, c1)
        try:
            os.remove(tgt)
        except OSError as e:
            if e.errno != errno.ENOENT:
                if not getattr(self, 'uninstall_error', None):
                    self.uninstall_error = True
                    Logs.warn('build: some files could not be uninstalled (retry with -vv to list them)')
                if Logs.verbose > 1:
                    Logs.warn('Could not remove %s (error code %r)', e.filename, e.errno)
        self.rm_empty_dirs(tgt)

    def do_unlink(self, src, tgt, **kw):
        try:
            if not self.generator.bld.progress_bar:
                c1 = Logs.colors.NORMAL
                c2 = Logs.colors.BLUE
                Logs.info('%s- remove %s%s%s', c1, c2, tgt, c1)
            os.remove(tgt)
        except OSError:
            pass
        self.rm_empty_dirs(tgt)


class InstallContext(BuildContext):
    '''installs the targets on the system'''
    cmd = 'install'

    def __init__(self, **kw):
        super(InstallContext, self).__init__(**kw)
        self.is_install = INSTALL


class UninstallContext(InstallContext):
    '''removes the targets installed'''
    cmd = 'uninstall'

    def __init__(self, **kw):
        super(UninstallContext, self).__init__(**kw)
        self.is_install = UNINSTALL


class CleanContext(BuildContext):
    '''cleans the project'''
    cmd = 'clean'

    def execute(self):
        self.restore()
        if not self.all_envs:
            self.load_envs()
        self.recurse([self.run_dir])
        try:
            self.clean()
        finally:
            self.store()

    def clean(self):
        Logs.debug('build: clean called')
        if hasattr(self, 'clean_files'):
            for n in self.clean_files:
                n.delete()
        elif self.bldnode != self.srcnode:
            lst = []
            for env in self.all_envs.values():
                lst.extend(self.root.find_or_declare(f) for f in env[CFG_FILES])
            excluded_dirs = '.lock* *conf_check_*/** config.log %s/*' % CACHE_DIR
            for n in self.bldnode.ant_glob('**/*', excl=excluded_dirs, quiet=True):
                if n in lst:
                    continue
                n.delete()
        self.root.children = {}
        for v in SAVED_ATTRS:
            if v == 'root':
                continue
            setattr(self, v, {})


class ListContext(BuildContext):
    '''lists the targets to execute'''
    cmd = 'list'

    def execute(self):
        self.restore()
        if not self.all_envs:
            self.load_envs()
        self.recurse([self.run_dir])
        self.pre_build()
        self.timer = Utils.Timer()
        for g in self.groups:
            for tg in g:
                try:
                    f = tg.post
                except AttributeError:
                    pass
                else:
                    f()
        try:
            self.get_tgen_by_name('')
        except Errors.WafError:
            pass
        targets = sorted(self.task_gen_cache_names)
        line_just = max(len(t) for t in targets) if targets else 0
        for target in targets:
            tgen = self.task_gen_cache_names[target]
            descript = getattr(tgen, 'description', '')
            if descript:
                target = target.ljust(line_just)
                descript = ': %s' % descript
            Logs.pprint('GREEN', target, label=descript)


class StepContext(BuildContext):
    '''executes tasks in a step-by-step fashion, for debugging'''
    cmd = 'step'

    def __init__(self, **kw):
        super(StepContext, self).__init__(**kw)
        self.files = Options.options.files

    def compile(self):
        if not self.files:
            Logs.warn('Add a pattern for the debug build, for example "waf step --files=main.c,app"')
            BuildContext.compile(self)
            return
        targets = []
        if self.targets and self.targets != '*':
            targets = self.targets.split(',')
        for g in self.groups:
            for tg in g:
                if targets and tg.name not in targets:
                    continue
                try:
                    f = tg.post
                except AttributeError:
                    pass
                else:
                    f()
            for pat in self.files.split(','):
                matcher = self.get_matcher(pat)
                for tg in g:
                    if isinstance(tg, Task.Task):
                        lst = [tg]
                    else:
                        lst = tg.tasks
                    for tsk in lst:
                        do_exec = False
                        for node in tsk.inputs:
                            if matcher(node, output=False):
                                do_exec = True
                                break
                        for node in tsk.outputs:
                            if matcher(node, output=True):
                                do_exec = True
                                break
                        if do_exec:
                            ret = tsk.run()
                            Logs.info('%s -> exit %r', tsk, ret)

    def get_matcher(self, pat):
        inn = True
        out = True
        if pat.startswith('in:'):
            out = False
            pat = pat.replace('in:', '')
        elif pat.startswith('out:'):
            inn = False
            pat = pat.replace('out:', '')
        anode = self.root.find_node(pat)
        pattern = None
        if not anode:
            if not pat.startswith('^'):
                pat = '^.+?%s' % pat
            if not pat.endswith('$'):
                pat = '%s$' % pat
            pattern = re.compile(pat)

        def match(node, output):
            if output and not out:
                return False
            if not output and not inn:
                return False
            if anode:
                return anode == node
            else:
                return pattern.match(node.abspath())

        return match


class EnvContext(BuildContext):
    fun = cmd = None

    def execute(self):
        self.restore()
        if not self.all_envs:
            self.load_envs()
        self.recurse([self.run_dir])



import os
import sys
import unittest

from PyInstaller.lib.modulegraph import modulegraph


# XXX: Todo: simular tests with bytecompiled modules


class TestEdgeData (unittest.TestCase):
    if not hasattr(unittest.TestCase, 'assertIsInstance'):
        def assertIsInstance(self, value, types):
            if not isinstance(value, types):
                self.fail("%r is not an instance of %r"%(value, types))

    def test_regular_import(self):
        root = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'testpkg-edgedata')
        mf = modulegraph.ModuleGraph(path=[ root ] + sys.path)
        script_name = os.path.join(root, 'script.py')
        mf.add_script(script_name)

        script_node = mf.find_node(script_name)
        self.assertIsInstance(script_node, modulegraph.Script)


        node = mf.find_node('toplevel_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=False, fromlist=False))

        node = mf.find_node('toplevel_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=False, fromlist=False))

        node = mf.find_node('toplevel_class_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=False, fromlist=False))

        node = mf.find_node('toplevel_class_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=False, fromlist=False))

        node = mf.find_node('toplevel_conditional_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=False, tryexcept=False, fromlist=False))

        node = mf.find_node('toplevel_conditional_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=False, tryexcept=False, fromlist=False))

        node = mf.find_node('toplevel_conditional_import_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=False, tryexcept=True, fromlist=False))

        node = mf.find_node('toplevel_conditional_import_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=False, tryexcept=True, fromlist=False))

        node = mf.find_node('toplevel_conditional_import2_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=False, tryexcept=True, fromlist=False))

        node = mf.find_node('toplevel_conditional_import2_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=False, tryexcept=True, fromlist=False))

        node = mf.find_node('toplevel_import_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=True, fromlist=False))

        node = mf.find_node('toplevel_import_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=True, fromlist=False))

        node = mf.find_node('toplevel_import2_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=True, fromlist=False))

        node = mf.find_node('toplevel_import2_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=True, fromlist=False))

        node = mf.find_node('function_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=False, fromlist=False))

        node = mf.find_node('function_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=False, fromlist=False))

        node = mf.find_node('function_class_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=False, fromlist=False))

        node = mf.find_node('function_class_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=False, fromlist=False))

        node = mf.find_node('function_conditional_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=True, tryexcept=False, fromlist=False))

        node = mf.find_node('function_conditional_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=True, tryexcept=False, fromlist=False))

        node = mf.find_node('function_conditional_import_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=True, tryexcept=True, fromlist=False))

        node = mf.find_node('function_conditional_import_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=True, tryexcept=True, fromlist=False))

        node = mf.find_node('function_conditional_import2_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=True, tryexcept=True, fromlist=False))

        node = mf.find_node('function_conditional_import2_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=True, tryexcept=True, fromlist=False))

        node = mf.find_node('function_import_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=True, fromlist=False))

        node = mf.find_node('function_import_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=True, fromlist=False))

        node = mf.find_node('function_import2_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=True, fromlist=False))

        node = mf.find_node('function_import2_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=True, fromlist=False))


    def test_multi_import(self):
        root = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'testpkg-edgedata')
        mf = modulegraph.ModuleGraph(path=[ root ] + sys.path)
        script_name = os.path.join(root, 'script_multi_import.py')
        mf.add_script(script_name)

        script_node = mf.find_node(script_name)
        self.assertIsInstance(script_node, modulegraph.Script)

        # FIXME PyInstaller: original _load_tail returned a MissingModule if
        # (_save)_import_module did return None. PyInstaller changed this in
        # cae47e4f5b51a94ac3ceb5d093283ba0cc895589 and raises an ImportError.
        # Thus the MissingModule node (which was expected to be created when
        # scanning the script above) doesn't exist and findNode will return
        # None, which makes this test fail.
        #node = mf.find_node('os.path')
        #ed = mf.edgeData(script_node, node)
        #self.assertIsInstance(ed, modulegraph.DependencyInfo)
        #self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=True, fromlist=False))

        node = mf.find_node('os')
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=False, fromlist=False))

        node = mf.find_node('sys')
        ed = mf.edgeData(script_node, node)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=True, tryexcept=False, fromlist=False))

        node = mf.find_node('platform')
        ed = mf.edgeData(script_node, node)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=True, tryexcept=False, fromlist=False))

        node = mf.find_node('email')
        ed = mf.edgeData(script_node, node)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=False, fromlist=False))

    def test_from_imports(self):
        root = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'testpkg-edgedata')
        mf = modulegraph.ModuleGraph(path=[ root ] + sys.path)
        script_name = os.path.join(root, 'script_from_import.py')
        mf.add_script(script_name)

        script_node = mf.find_node(script_name)
        self.assertIsInstance(script_node, modulegraph.Script)


        node = mf.find_node('pkg.toplevel_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=False, fromlist=True))

        node = mf.find_node('pkg.toplevel_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=False, fromlist=True))

        node = mf.find_node('pkg.toplevel_class_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=False, fromlist=True))

        node = mf.find_node('pkg.toplevel_class_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=False, fromlist=True))

        node = mf.find_node('pkg.toplevel_conditional_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=False, tryexcept=False, fromlist=True))

        node = mf.find_node('pkg.toplevel_conditional_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=False, tryexcept=False, fromlist=True))

        node = mf.find_node('pkg.toplevel_conditional_import_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=False, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.toplevel_conditional_import_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=False, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.toplevel_conditional_import2_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=False, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.toplevel_conditional_import2_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=False, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.toplevel_import_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.toplevel_import_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.toplevel_import2_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.toplevel_import2_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=False, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.function_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=False, fromlist=True))

        node = mf.find_node('pkg.function_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=False, fromlist=True))

        node = mf.find_node('pkg.function_class_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=False, fromlist=True))

        node = mf.find_node('pkg.function_class_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=False, fromlist=True))

        node = mf.find_node('pkg.function_conditional_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=True, tryexcept=False, fromlist=True))

        node = mf.find_node('pkg.function_conditional_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=True, tryexcept=False, fromlist=True))

        node = mf.find_node('pkg.function_conditional_import_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=True, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.function_conditional_import_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=True, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.function_conditional_import2_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=True, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.function_conditional_import2_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=True, function=True, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.function_import_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.function_import_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.function_import2_existing')
        self.assertIsInstance(node, modulegraph.SourceModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=True, fromlist=True))

        node = mf.find_node('pkg.function_import2_nonexisting')
        self.assertIsInstance(node, modulegraph.MissingModule)
        ed = mf.edgeData(script_node, node)
        self.assertIsInstance(ed, modulegraph.DependencyInfo)
        self.assertEqual(ed, modulegraph.DependencyInfo(conditional=False, function=True, tryexcept=True, fromlist=True))


if __name__ == "__main__":
    unittest.main()



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
macOS-specific test to check handling of Apple Events in the bootloader.
"""

import json
import os
import subprocess
import time
import functools
import uuid

import pytest

from PyInstaller.compat import is_macos_11


# On macOS 11, the custom URL schema registration does not work properly if the .app bundle is located in the default
# temporary path (/var or /private/var prefix). Therefore, for the tests below to work, the pytest's base temporary
# path needs to be moved (for example, via the --basetemp argument).
def macos11_check_tmp_path(test):
    @functools.wraps(test)
    def wrapped(**kwargs):
        tmp_path = kwargs['tmp_path']
        if is_macos_11 and str(tmp_path).startswith(('/var', '/private/var')):
            pytest.skip(
                "The custom URL schema registration does not work on macOS 11 when .app bundles are placed in "
                "the default temporary path."
            )
        return test(**kwargs)

    return wrapped


# This test function is similar to the former test_osx_event_forwarding, but is designed to test onefile vs onedir
# builds, with argv_emu on and off. It is also designed to be used with different Apple Event loggers based on
# different toolkits: Carbon via ctypes, tkinter, PyQt5, provided the logger implements the common log file format.
#
# However, in practice, in turns out that each high-level toolkit handles events a bit differently, and offers different
# level of support (in tkinter/Tk case even depending on the Tk version). Therefore, this test function is currently
# used only with Carbon-based logger and does not (yet) implement toolkit-specific quirks.
def _test_apple_events_handling(appname, tmp_path, pyi_builder_spec, monkeypatch, build_mode, argv_emu):
    # Helper for determining application start/finish via logged events.
    def wait_for_event(logfile, event, timeout=60, polltime=0.25):
        """
        Wait for the log file with 'started' or 'finished' entry to appear.
        """
        assert event in {'started', 'finished'}, f"Invalid event: {event}!"
        t0 = time.time()  # mark start time
        # Poll logfile for desired event (the app writes 'started' at beginning and 'finished' at the end).
        while True:
            elapsed = time.time() - t0
            if elapsed > timeout:
                return False
            if os.path.exists(logfile_path):
                with open(logfile_path, 'r', encoding='utf8') as fh:
                    log_lines = fh.readlines()
                    if log_lines:
                        if event == 'started':
                            first = log_lines[0]
                            assert first.startswith('started '), "Unexpected line in log file!"
                            return True
                        elif log_lines[-1].startswith('finished '):
                            return True
            # Try again later
            time.sleep(polltime)

    # Generate unique URL scheme & file ext to avoid collisions.
    unique_key = uuid.uuid4()
    custom_url_scheme = f"pyi-test-{unique_key}"
    custom_file_ext = f"pyi_test_{unique_key}"
    monkeypatch.setenv("PYI_CUSTOM_URL_SCHEME", custom_url_scheme)
    monkeypatch.setenv("PYI_CUSTOM_FILE_EXT", custom_file_ext)
    monkeypatch.setenv("PYI_BUILD_MODE", build_mode)
    monkeypatch.setenv("PYI_ARGV_EMU", str(int(argv_emu)))

    logfile_path = tmp_path / 'dist' / 'events.log'

    # test_spec() builds the app and automatically runs it. As we want to test custom protocol handler registration, we
    # want this run to exit as soon as possible. Passing arg "0" would have it exit immediately, but with some logger
    # applications (e.g., pure Carbon based), that prevents bundle's executable to be found on subsequent open attempts;
    # presumably because event loop does not process any events. So use 5-seconds timeout instead.
    pyi_builder_spec.test_spec(f'{appname}.spec', app_args=["5"])

    # Wait for the app started by test_spec() to exit
    assert wait_for_event(logfile_path, 'started'), 'Timeout while waiting for app to start (test_spec run)!'
    assert wait_for_event(logfile_path, 'finished'), 'Timeout while waiting for app to finish (test_spec run)!'
    time.sleep(5)  # wait for app to fully exit

    # Clean up the log file created by test_spec() running the app
    os.remove(logfile_path)

    # Rename the dist directory into dist-{uuid}, to ensure path uniqueness for each test run. The name of the temporary
    # directory may be the same across different test runs (with different parametrizations) due to the length of the
    # test name; re-using the same path (even though the preceding test's contents were removed) may cause issues with
    # app bundle registration...
    old_dist = tmp_path / 'dist'
    new_dist = tmp_path / f'dist-{unique_key}'

    old_dist.rename(new_dist)

    app_path = new_dist / f'{appname}.app'
    logfile_path = new_dist / 'events.log'

    # Run using 'open', passing a 5-second timeout as an arg to exit as soon as possible (do not pass 0 to prevent
    # skipping the event loop in the application). This will cause macOS to register the custom protocol handler and
    # file extension association.
    subprocess.check_call(['open', str(app_path), '--args', "5"])

    assert wait_for_event(logfile_path, 'started'), 'Timeout while waiting for app to start (registration run)!'
    assert wait_for_event(logfile_path, 'finished'), 'Timeout while waiting for app to finish (registration run)!'
    time.sleep(5)  # wait for app to fully exit

    # App exited immediately, clean-up
    logfile_path.unlink()

    # At this point both the protocol handler and the file ext are registered
    # 1. Try the file extension -- this tests the AppleEvent rewrite of a "file://" event to a regular filesystem path.

    # Create 32 files that are associated with this app. This tests the robustness of the argv-emu by spamming it with
    # lots of args and seeing what happens.
    n_files = 32
    assoc_files = []
    for idx in range(n_files):
        assoc_path = tmp_path / f'AFile{idx}.{custom_file_ext}'
        assoc_path.write_text(f"File contents #{idx}\n", encoding='utf-8')
        assoc_files.append(str(assoc_path))

    # Open app again by "open"ing the associated files.
    #
    # These are sent as Apple Events to the app immediately after it starts; if argv emulation is enabled, the
    # bootloader processes these events and translates them into file paths, passing them as argv to the
    # subordinate app.
    #
    # The generator below produces odd numbered files as "file://" URLs, and even numbered are just file paths. Under
    # argv emulation, they should all end up appended to sys.argv in the app as simple file paths.
    files_list = [('file://' if idx % 2 else '') + file_path for idx, file_path in enumerate(assoc_files)]
    subprocess.check_call(['open', *files_list])

    assert wait_for_event(logfile_path, 'started'), 'Timeout while waiting for app to start (test run)!'

    # At this point the app is running.

    # 2. Call open using the url associated with the app. This should forward the Apple URL event to the running app.
    url_hello = f"{custom_url_scheme}://lowecase_required/hello_world/"
    subprocess.check_call(['open', url_hello])
    time.sleep(1.0)

    # 3. Put application into background so that we can re-activate it. Taken from the former test_osx_event_forwarding
    # that used PyQt5, in case we ever try to use this test function with PyQt5-based logger.
    app_put_to_background = False
    try:
        # On GitHub Actions macos-11 runner, launching osascript sometimes fails with the following error:
        # "osascript: can't open default scripting component."
        subprocess.check_call(['osascript', "-e", 'tell application "System Events" to activate'])
        app_put_to_background = True
    except Exception:
        pass
    time.sleep(1.0)  # delay for above applescript

    if app_put_to_background:
        # The app is running now, in the background, and does not have focus.

        # 4. Call open passing the app path again -- this should activate the already-running app.
        subprocess.check_call(['open', app_path])
        time.sleep(1.0)  # the activate event gets sent with a delay

    # 5. Call open with first four associated files.
    files_list = [('file://' if idx % 2 else '') + file_path for idx, file_path in enumerate(assoc_files[:4])]
    subprocess.check_call(['open', *files_list])
    time.sleep(1.0)

    # 6. Call open again using the url associated with the app. This should forward the Apple URL event to the
    # already-running app.
    url_goodbye = f"{custom_url_scheme}://lowecase_required/goodybe_galaxy/"
    subprocess.check_call(['open', url_goodbye])
    time.sleep(1.0)

    # 7. Call open again using the url associated with the app, this time testing support for large URL data (~64kB).
    url_large = f"{custom_url_scheme}://lowecase_required/large_data/"
    # Note: We would have gone larger but 'open' itself seems to not consistently like data over a certain size.
    url_large += 'x' * 64000  # Append 64 kB of data to URL to stress-test.
    subprocess.check_call(['open', url_large])
    time.sleep(1.0)

    # 8. Call open with last four associated files again.
    files_list = [('file://' if idx % 2 else '') + file_path for idx, file_path in enumerate(assoc_files[-4:])]
    subprocess.check_call(['open', *files_list])
    time.sleep(1.0)

    # 9. Wait for application to finish.
    assert wait_for_event(logfile_path, 'finished'), 'Timeout while waiting for app to finish (test run)!'
    time.sleep(2)  # wait for app to fully exit

    # *** Analyze the contents of the log file ***
    with open(logfile_path, 'r', encoding='utf8') as fh:
        log_lines = fh.readlines()

    assert log_lines[0].startswith('started '), "Unexpected first line in log!"
    assert log_lines[-1].startswith('finished '), "Unexpected last line in log!"

    events = []
    errors = []
    unknown = []
    for log_line in log_lines[1:-1]:
        if log_line.startswith('ae '):
            # (raw) Apple Event
            _, event_id, event_data = log_line.split(" ", 2)
            events.append((event_id, json.loads(event_data)))
        elif log_line.startswith('ERROR '):
            errors.append(log_line.split(" ", 1))
        else:
            unknown.append(log_line)

    assert not errors, "Event log contains error(s)!"
    assert not unknown, "Event log contains unknown line(s)!"

    # First line: application start - read arguments.
    data = json.loads(log_lines[0].split(" ", 1)[-1])
    args = data['args']  # Validated below, with initial event.

    # Last line: application finish - read activation count.
    data = json.loads(log_lines[-1].split(" ", 1)[-1])
    activation_count = data['activation_count']

    # Event: initial event
    #
    # In onefile builds, we always receive oapp event, regardless of how application was opened (e.g., via odoc or
    # GURL), because the actual opening event is received by the parent process, whereas child is independent and
    # is always launched via "normal" open (oapp). Therefore, having the argv-emu enabled or not also has no effect on
    # the initial oapp event in onefile mode.
    #
    # In onedir mode without argv emulation, we receive either oapp for normal open, or odoc/GURL if launched via open
    # document/URL. The test run begins with an open-file request, so the initial event should be an odoc one.
    #
    # When argv-emu is enabled in onedir mode, the argv-emu event loop swallows that initial event, whatever it may have
    # been. To make up for it and prevent issues with some toolkits (e.g., Tcl/Tk), it emits a fake oapp event.
    initial_oapp = True  # expect initial oapp event
    if build_mode == 'onedir' and not argv_emu:
        initial_oapp = False  # initial event is odoc

    event_idx = 0

    if initial_oapp:
        event, data = events[event_idx]
        event_idx += 1
        assert event == 'oapp'

    # Validate arguments received at application start via argv.
    if argv_emu:
        # argv-emu; we expect whole list of files in arguments
        assert args == assoc_files, "Arguments received via argv-emu do not match expected list!"
    else:
        # No argv-emu; args should be empty (in particular, the -psn_* arg should be filtered out).
        assert args == [], "Application should receive no arguments when argv-emu is disabled!"

        # The open-file request should be received as odoc request instead. Might be very first event, or might follow
        # an oapp, depending on the build_mode.
        event, data = events[event_idx]
        event_idx += 1
        assert event == 'odoc'
        assert data == assoc_files

    # Validate activation count at the time of finish. One contribution should be from the initial oapp event (if
    # applicable), and one from re-activation when we sent application to background and then called it forward
    # (provided there was no osascript error).
    expected_activations = initial_oapp + app_put_to_background
    assert activation_count == expected_activations, "Application did not handle activation event(s) as expected!"

    # The rest of events should be received normally, regardless of argv-emu. Either directly (onedir), or via the
    # forwarding mechanism (onefile).

    # Event: GURL with single URL (url_hello)
    event, data = events[event_idx]
    event_idx += 1
    assert event == 'GURL'
    assert data == [url_hello]

    # Event: rapp (re-activation)
    # Applicable only if we successfully put application to background via osascript
    if app_put_to_background:
        event, data = events[event_idx]
        event_idx += 1
        assert event == 'rapp'

    # Event: odoc with first 4 files
    event, data = events[event_idx]
    event_idx += 1
    assert event == 'odoc'
    assert data == assoc_files[:4]

    # Event: GURL with single URL (url_goodbye)
    event, data = events[event_idx]
    event_idx += 1
    assert event == 'GURL'
    assert data == [url_goodbye]

    # Event: GURL with single large-data URL (url_large)
    event, data = events[event_idx]
    event_idx += 1
    assert event == 'GURL'
    assert data == [url_large]

    # Event: odoc with last 4 files
    event, data = events[event_idx]
    event_idx += 1
    assert event == 'odoc'
    assert data == assoc_files[-4:]

    # *** Cleanup ***
    # Delete all the temp files to be polite
    for ff in assoc_files:
        try:
            os.remove(ff)
        except OSError:
            pass


@pytest.mark.darwin
@macos11_check_tmp_path
@pytest.mark.flaky(reruns=3)
@pytest.mark.parametrize("build_mode", ['onefile', 'onedir'])
@pytest.mark.parametrize("argv_emu", [True, False], ids=["emu", "noemu"])
def test_apple_event_handling_carbon(tmp_path, pyi_builder_spec, monkeypatch, build_mode, argv_emu):
    # Carbon-based event logger.
    return _test_apple_events_handling(
        'pyi_osx_aevent_handling_carbon',
        tmp_path,
        pyi_builder_spec,
        monkeypatch,
        build_mode,
        argv_emu,
    )



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from __future__ import with_statement
import os, re
from waflib import Context, Task, Utils, Logs
import waflib.Tools.ccroot
from waflib.TaskGen import feature, before_method, taskgen_method
from waflib.Logs import error
from waflib.Configure import conf

_style_flags = {
    'ba': '-b',
    'desktop': '-d',
    'keys': '-k',
    'quoted': '--quoted-style',
    'quotedxml': '--quotedxml-style',
    'rfc822deb': '-r',
    'schemas': '-s',
    'xml': '-x',
}


@taskgen_method
def ensure_localedir(self):
    if not self.env.LOCALEDIR:
        if self.env.DATAROOTDIR:
            self.env.LOCALEDIR = os.path.join(self.env.DATAROOTDIR, 'locale')
        else:
            self.env.LOCALEDIR = os.path.join(self.env.PREFIX, 'share', 'locale')


@before_method('process_source')
@feature('intltool_in')
def apply_intltool_in_f(self):
    try:
        self.meths.remove('process_source')
    except ValueError:
        pass
    self.ensure_localedir()
    podir = getattr(self, 'podir', '.')
    podirnode = self.path.find_dir(podir)
    if not podirnode:
        error("could not find the podir %r" % podir)
        return
    cache = getattr(self, 'intlcache', '.intlcache')
    self.env.INTLCACHE = [os.path.join(str(self.path.get_bld()), podir, cache)]
    self.env.INTLPODIR = podirnode.bldpath()
    self.env.append_value('INTLFLAGS', getattr(self, 'flags', self.env.INTLFLAGS_DEFAULT))
    if '-c' in self.env.INTLFLAGS:
        self.bld.fatal('Redundant -c flag in intltool task %r' % self)
    style = getattr(self, 'style', None)
    if style:
        try:
            style_flag = _style_flags[style]
        except KeyError:
            self.bld.fatal('intltool_in style "%s" is not valid' % style)
        self.env.append_unique('INTLFLAGS', [style_flag])
    for i in self.to_list(self.source):
        node = self.path.find_resource(i)
        task = self.create_task('intltool', node, node.change_ext(''))
        inst = getattr(self, 'install_path', None)
        if inst:
            self.add_install_files(install_to=inst, install_from=task.outputs)


@feature('intltool_po')
def apply_intltool_po(self):
    try:
        self.meths.remove('process_source')
    except ValueError:
        pass
    self.ensure_localedir()
    appname = getattr(self, 'appname', getattr(Context.g_module, Context.APPNAME, 'set_your_app_name'))
    podir = getattr(self, 'podir', '.')
    inst = getattr(self, 'install_path', '${LOCALEDIR}')
    linguas = self.path.find_node(os.path.join(podir, 'LINGUAS'))
    if linguas:
        with open(linguas.abspath()) as f:
            langs = []
            for line in f.readlines():
                if not line.startswith('#'):
                    langs += line.split()
        re_linguas = re.compile('[-a-zA-Z_@.]+')
        for lang in langs:
            if re_linguas.match(lang):
                node = self.path.find_resource(os.path.join(podir, re_linguas.match(lang).group() + '.po'))
                task = self.create_task('po', node, node.change_ext('.mo'))
                if inst:
                    filename = task.outputs[0].name
                    (langname, ext) = os.path.splitext(filename)
                    inst_file = inst + os.sep + langname + os.sep + 'LC_MESSAGES' + os.sep + appname + '.mo'
                    self.add_install_as(
                        install_to=inst_file, install_from=task.outputs[0], chmod=getattr(self, 'chmod', Utils.O644)
                    )
    else:
        Logs.pprint('RED', "Error no LINGUAS file found in po directory")


class po(Task.Task):
    run_str = '${MSGFMT} -o ${TGT} ${SRC}'
    color = 'BLUE'


class intltool(Task.Task):
    run_str = '${INTLTOOL} ${INTLFLAGS} ${INTLCACHE_ST:INTLCACHE} ${INTLPODIR} ${SRC} ${TGT}'
    color = 'BLUE'


@conf
def find_msgfmt(conf):
    conf.find_program('msgfmt', var='MSGFMT')


@conf
def find_intltool_merge(conf):
    if not conf.env.PERL:
        conf.find_program('perl', var='PERL')
    conf.env.INTLCACHE_ST = '--cache=%s'
    conf.env.INTLFLAGS_DEFAULT = ['-q', '-u']
    conf.find_program('intltool-merge', interpreter='PERL', var='INTLTOOL')


def configure(conf):
    conf.find_msgfmt()
    conf.find_intltool_merge()
    if conf.env.CC or conf.env.CXX:
        conf.check(header_name='locale.h')



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib.TaskGen import extension
from waflib import Task


@extension('.lua')
def add_lua(self, node):
    tsk = self.create_task('luac', node, node.change_ext('.luac'))
    inst_to = getattr(self, 'install_path', self.env.LUADIR and '${LUADIR}' or None)
    if inst_to:
        self.add_install_files(install_to=inst_to, install_from=tsk.outputs)
    return tsk


class luac(Task.Task):
    run_str = '${LUAC} -s -o ${TGT} ${SRC}'
    color = 'PINK'


def configure(conf):
    conf.find_program('luac', var='LUAC')



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Code to be shared by PyInstaller and the bootloader/wscript file.

This code must not assume that either PyInstaller or any of its dependencies installed. I.e., the only imports allowed
in here are standard library ones. Within reason, it is preferable that this file should still run under Python 2.7 as
many compiler docker images still have only Python 2 installed.
"""

import platform
import re


def _pyi_machine(machine, system):
    # type: (str, str) -> str
    """
    Choose an intentionally simplified architecture identifier to be used in the bootloader's directory name.

    Args:
        machine:
            The output of ``platform.machine()`` or any known architecture alias or shorthand that may be used by a
            C compiler.
        system:
            The output of ``platform.system()`` on the target machine.
    Returns:
        Either a string tag or, on platforms that don't need an architecture tag, ``None``.

    Ideally, we would just use ``platform.machine()`` directly, but that makes cross-compiling the bootloader almost
    impossible, because you need to know at compile time exactly what ``platform.machine()`` will be at run time, based
    only on the machine name alias or shorthand reported by the C compiler at the build time. Rather, use a loose
    differentiation, and trust that anyone mixing armv6l with armv6h knows what they are doing.
    """
    # See the corresponding tests in tests/unit/test_compat.py for examples.

    if platform.machine() == "sw_64" or platform.machine() == "loongarch64":
        # This explicitly inhibits cross compiling the bootloader for or on SunWay and LoongArch machine.
        return platform.machine()

    if system == "Windows":
        if machine.lower().startswith("arm"):
            return "arm"
        else:
            return "intel"

    if system == "SunOS":
        if machine.lower() in ("x86_64", "x86", "i86pc"):
            return "intel"
        else:
            return "sparc"

    if system != "Linux":
        # No architecture specifier for anything par Linux.
        # - macOS is on two 64 bit architectures, but they are merged into one "universal2" bootloader.
        # - BSD supports a wide range of architectures, but according to PyPI's download statistics, every one of our
        #   BSD users are on x86_64. This may change in the distant future.
        return

    if machine.startswith(("arm", "aarch")):
        # ARM has a huge number of similar and aliased sub-versions, such as armv5, armv6l armv8h, aarch64.
        return "arm"
    if machine in ("thumb"):
        # Reported by waf/gcc when Thumb instruction set is enabled on 32-bit ARM. The platform.machine() returns "arm"
        # regardless of the instruction set.
        return "arm"
    if machine in ("x86_64", "x64", "x86"):
        return "intel"
    if re.fullmatch("i[1-6]86", machine):
        return "intel"
    if machine.startswith(("ppc", "powerpc")):
        # PowerPC comes in 64 vs 32 bit and little vs big endian variants.
        return "ppc"
    if machine in ("mips64", "mips"):
        return "mips"
    if machine.startswith("riscv"):
        return "riscv"
    # Machines with no known aliases :)
    if machine in ("s390x",):
        return machine

    # Unknown architectures are allowed by default, but will all be placed under one directory. In theory, trying to
    # have multiple unknown architectures in one copy of PyInstaller will not work, but that should be sufficiently
    # unlikely to ever happen.
    return "unknown"
